import json
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple
from uuid import UUID

import httpx
from httpx import HTTPError, HTTPStatusError
from loguru import logger

from langflow.services.base import Service
from langflow.services.store.exceptions import APIKeyError, FilterError, ForbiddenError
from langflow.services.store.schema import (
    CreateComponentResponse,
    DownloadComponentResponse,
    ListComponentResponse,
    ListComponentResponseModel,
    StoreComponentCreate,
)
from langflow.services.store.utils import (
    process_component_data,
    process_tags_for_post,
    update_components_with_user_data,
)

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService

from contextlib import asynccontextmanager
from contextvars import ContextVar

user_data_var: ContextVar[Optional[Dict[str, Any]]] = ContextVar("user_data", default=None)


@asynccontextmanager
async def user_data_context(store_service: "StoreService", api_key: Optional[str] = None):
    # Fetch and set user data to the context variable
    if api_key:
        try:
            user_data, _ = await store_service._get(
                f"{store_service.base_url}/users/me", api_key, params={"fields": "id"}
            )
            user_data_var.set(user_data[0])
        except HTTPStatusError as exc:
            if exc.response.status_code == 403:
                raise ValueError("Invalid API key")
    try:
        yield
    finally:
        # Clear the user data from the context variable
        user_data_var.set(None)


class StoreService(Service):
    """This is a service that integrates langflow with the store which
    is a Directus instance. It allows to search, get and post components to
    the store."""

    name = "store_service"

    def __init__(self, settings_service: "SettingsService"):
        self.settings_service = settings_service
        self.base_url = self.settings_service.settings.STORE_URL
        self.download_webhook_url = self.settings_service.settings.DOWNLOAD_WEBHOOK_URL
        self.like_webhook_url = self.settings_service.settings.LIKE_WEBHOOK_URL
        self.components_url = f"{self.base_url}/items/components"
        self.default_fields = [
            "id",
            "name",
            "description",
            "user_created.username",
            "is_component",
            "tags.tags_id.name",
            "tags.tags_id.id",
            "count(liked_by)",
            "count(downloads)",
            "metadata",
            "last_tested_version",
            "private",
        ]

    # Create a context manager that will use the api key to
    # get the user data and all requests inside the context manager
    # will make a property return that data
    # Without making the request multiple times

    async def check_api_key(self, api_key: str):
        # Check if the api key is valid
        # If it is, return True
        # If it is not, return False
        try:
            user_data, _ = await self._get(f"{self.base_url}/users/me", api_key, params={"fields": "id"})

            return "id" in user_data[0]
        except HTTPStatusError as exc:
            if exc.response.status_code in [403, 401]:
                return False
            else:
                raise ValueError(f"Unexpected status code: {exc.response.status_code}")
        except Exception as exc:
            raise ValueError(f"Unexpected error: {exc}")

    async def _get(
        self, url: str, api_key: Optional[str] = None, params: Optional[Dict[str, Any]] = None
    ) -> Tuple[List[Dict[str, Any]], Dict[str, Any]]:
        """Utility method to perform GET requests."""
        if api_key:
            headers = {"Authorization": f"Bearer {api_key}"}
        else:
            headers = {}
        async with httpx.AsyncClient() as client:
            try:
                response = await client.get(url, headers=headers, params=params)
                response.raise_for_status()
            except HTTPError as exc:
                raise exc
            except Exception as exc:
                raise ValueError(f"GET failed: {exc}")
        json_response = response.json()
        result = json_response["data"]
        metadata = {}
        if "meta" in json_response:
            metadata = json_response["meta"]

        if isinstance(result, dict):
            return [result], metadata
        return result, metadata

    async def call_webhook(self, api_key: str, webhook_url: str, component_id: UUID) -> None:
        # The webhook is a POST request with the data in the body
        # For now we are calling it just for testing
        try:
            headers = {"Authorization": f"Bearer {api_key}"}
            async with httpx.AsyncClient() as client:
                response = await client.post(webhook_url, headers=headers, json={"component_id": str(component_id)})
                response.raise_for_status()
            return response.json()
        except HTTPError as exc:
            raise exc
        except Exception as exc:
            logger.debug(f"Webhook failed: {exc}")

    def build_tags_filter(self, tags: List[str]):
        tags_filter: Dict[str, Any] = {"tags": {"_and": []}}
        for tag in tags:
            tags_filter["tags"]["_and"].append({"_some": {"tags_id": {"name": {"_eq": tag}}}})
        return tags_filter

    async def count_components(
        self,
        filter_conditions: List[Dict[str, Any]],
        api_key: Optional[str] = None,
        use_api_key: Optional[bool] = False,
    ) -> int:
        params = {"aggregate": json.dumps({"count": "*"})}
        if filter_conditions:
            params["filter"] = json.dumps({"_and": filter_conditions})

        api_key = api_key if use_api_key else None

        results, _ = await self._get(self.components_url, api_key, params)
        return int(results[0].get("count", 0))

    @staticmethod
    def build_search_filter_conditions(query: str):
        # instead of build the param ?search=query, we will build the filter
        # that will use _icontains (case insensitive)
        conditions: Dict[str, Any] = {"_or": []}
        conditions["_or"].append({"name": {"_icontains": query}})
        conditions["_or"].append({"description": {"_icontains": query}})
        conditions["_or"].append({"tags": {"tags_id": {"name": {"_icontains": query}}}})
        conditions["_or"].append({"user_created": {"username": {"_icontains": query}}})
        return conditions

    def build_filter_conditions(
        self,
        component_id: Optional[str] = None,
        search: Optional[str] = None,
        private: Optional[bool] = None,
        tags: Optional[List[str]] = None,
        is_component: Optional[bool] = None,
        filter_by_user: Optional[bool] = False,
        liked: Optional[bool] = False,
        store_api_key: Optional[str] = None,
    ):
        filter_conditions = []

        if search is not None:
            search_conditions = self.build_search_filter_conditions(search)
            filter_conditions.append(search_conditions)

        if private is not None:
            filter_conditions.append({"private": {"_eq": private}})

        if tags:
            tags_filter = self.build_tags_filter(tags)
            filter_conditions.append(tags_filter)
        if component_id is not None:
            filter_conditions.append({"id": {"_eq": component_id}})
        if is_component is not None:
            filter_conditions.append({"is_component": {"_eq": is_component}})
        if liked and store_api_key:
            liked_filter = self.build_liked_filter()
            filter_conditions.append(liked_filter)
        elif liked and not store_api_key:
            raise APIKeyError("You must provide an API key to filter by likes")

        if filter_by_user and store_api_key:
            user_data = user_data_var.get()
            if not user_data:
                raise ValueError("No user data")
            filter_conditions.append({"user_created": {"_eq": user_data["id"]}})
        elif filter_by_user and not store_api_key:
            raise APIKeyError("You must provide an API key to filter your components")
        else:
            filter_conditions.append({"private": {"_eq": False}})

        return filter_conditions

    def build_liked_filter(self):
        user_data = user_data_var.get()
        # params["filter"] = json.dumps({"user_created": {"_eq": user_data["id"]}})
        if not user_data:
            raise ValueError("No user data")
        return {"liked_by": {"directus_users_id": {"_eq": user_data["id"]}}}

    async def query_components(
        self,
        api_key: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page: int = 1,
        limit: int = 15,
        fields: Optional[List[str]] = None,
        filter_conditions: Optional[List[Dict[str, Any]]] = None,
        use_api_key: Optional[bool] = False,
    ) -> Tuple[List[ListComponentResponse], Dict[str, Any]]:
        params: Dict[str, Any] = {
            "page": page,
            "limit": limit,
            "fields": ",".join(fields) if fields is not None else ",".join(self.default_fields),
            "meta": "filter_count",  # !This is DEPRECATED so we should remove it ASAP
        }
        # ?aggregate[count]=likes

        if sort:
            params["sort"] = ",".join(sort)

        # Only public components or the ones created by the user
        # check for "public" or "Public"

        if filter_conditions:
            params["filter"] = json.dumps({"_and": filter_conditions})

        # If not liked, this means we are getting public components
        # so we don't need to risk passing an invalid api_key
        # and getting 401
        api_key = api_key if use_api_key else None
        results, metadata = await self._get(self.components_url, api_key, params)
        if isinstance(results, dict):
            results = [results]

        results_objects = [ListComponentResponse(**result) for result in results]

        return results_objects, metadata

    async def get_liked_by_user_components(self, component_ids: List[str], api_key: str) -> List[str]:
        # Get fields id
        # filter should be "id is in component_ids AND liked_by directus_users_id token is api_key"
        # return the ids
        user_data = user_data_var.get()
        if not user_data:
            raise ValueError("No user data")
        params = {
            "fields": "id",
            "filter": json.dumps(
                {
                    "_and": [
                        {"id": {"_in": component_ids}},
                        {"liked_by": {"directus_users_id": {"_eq": user_data["id"]}}},
                    ]
                }
            ),
        }
        results, _ = await self._get(self.components_url, api_key, params)
        return [result["id"] for result in results]

    # Which of the components is parent of the user's components
    async def get_components_in_users_collection(self, component_ids: List[str], api_key: str):
        user_data = user_data_var.get()
        if not user_data:
            raise ValueError("No user data")
        params = {
            "fields": "id",
            "filter": json.dumps(
                {
                    "_and": [
                        {"user_created": {"_eq": user_data["id"]}},
                        {"parent": {"_in": component_ids}},
                    ]
                }
            ),
        }
        results, _ = await self._get(self.components_url, api_key, params)
        return [result["id"] for result in results]

    async def download(self, api_key: str, component_id: UUID) -> DownloadComponentResponse:
        url = f"{self.components_url}/{component_id}"
        params = {"fields": ",".join(["id", "name", "description", "data", "is_component", "metadata"])}
        if not self.download_webhook_url:
            raise ValueError("DOWNLOAD_WEBHOOK_URL is not set")
        component, _ = await self._get(url, api_key, params)
        await self.call_webhook(api_key, self.download_webhook_url, component_id)
        if len(component) > 1:
            raise ValueError("Something went wrong while downloading the component")
        component_dict = component[0]

        download_component = DownloadComponentResponse(**component_dict)
        # Check if metadata is an empty dict
        if download_component.metadata in [None, {}] and download_component.data is not None:
            # If it is, we need to build the metadata
            try:
                download_component.metadata = process_component_data(download_component.data.get("nodes", []))
            except KeyError:
                raise ValueError("Invalid component data. No nodes found")
        return download_component

    async def upload(self, api_key: str, component_data: StoreComponentCreate) -> CreateComponentResponse:
        headers = {"Authorization": f"Bearer {api_key}"}
        component_dict = component_data.model_dump(exclude_unset=True)
        # Parent is a UUID, but the store expects a string
        response = None
        if component_dict.get("parent"):
            component_dict["parent"] = str(component_dict["parent"])

        component_dict = process_tags_for_post(component_dict)
        try:
            # response = httpx.post(self.components_url, headers=headers, json=component_dict)
            # response.raise_for_status()
            async with httpx.AsyncClient() as client:
                response = await client.post(self.components_url, headers=headers, json=component_dict)
                response.raise_for_status()
            component = response.json()["data"]
            return CreateComponentResponse(**component)
        except HTTPError as exc:
            if response:
                try:
                    errors = response.json()
                    message = errors["errors"][0]["message"]
                    if message == "An unexpected error occurred.":
                        # This is a bug in Directus that returns this error
                        # when an error was thrown in the flow
                        message = "You already have a component with this name. Please choose a different name."
                    raise FilterError(message)
                except UnboundLocalError:
                    pass
            raise ValueError(f"Upload failed: {exc}")

    async def update(
        self, api_key: str, component_id: UUID, component_data: StoreComponentCreate
    ) -> CreateComponentResponse:
        # Patch is the same as post, but we need to add the id to the url
        headers = {"Authorization": f"Bearer {api_key}"}
        component_dict = component_data.model_dump(exclude_unset=True)
        # Parent is a UUID, but the store expects a string
        response = None
        if component_dict.get("parent"):
            component_dict["parent"] = str(component_dict["parent"])

        component_dict = process_tags_for_post(component_dict)
        try:
            # response = httpx.post(self.components_url, headers=headers, json=component_dict)
            # response.raise_for_status()
            async with httpx.AsyncClient() as client:
                response = await client.patch(
                    self.components_url + f"/{component_id}", headers=headers, json=component_dict
                )
                response.raise_for_status()
            component = response.json()["data"]
            return CreateComponentResponse(**component)
        except HTTPError as exc:
            if response:
                try:
                    errors = response.json()
                    message = errors["errors"][0]["message"]
                    if message == "An unexpected error occurred.":
                        # This is a bug in Directus that returns this error
                        # when an error was thrown in the flow
                        message = "You already have a component with this name. Please choose a different name."
                    raise FilterError(message)
                except UnboundLocalError:
                    pass
            raise ValueError(f"Upload failed: {exc}")

    async def get_tags(self) -> List[Dict[str, Any]]:
        url = f"{self.base_url}/items/tags"
        params = {"fields": ",".join(["id", "name"])}
        tags, _ = await self._get(url, api_key=None, params=params)
        return tags

    async def get_user_likes(self, api_key: str) -> List[Dict[str, Any]]:
        url = f"{self.base_url}/users/me"
        params = {
            "fields": ",".join(["id", "likes"]),
        }
        likes, _ = await self._get(url, api_key, params)
        return likes

    async def get_component_likes_count(self, component_id: str, api_key: Optional[str] = None) -> int:
        url = f"{self.components_url}/{component_id}"

        params = {
            "fields": ",".join(["id", "count(liked_by)"]),
        }
        result, _ = await self._get(url, api_key=api_key, params=params)
        if len(result) == 0:
            raise ValueError("Component not found")
        likes = result[0]["liked_by_count"]
        # likes_by_count is a string
        # try to convert it to int
        try:
            likes = int(likes)
        except ValueError:
            raise ValueError(f"Unexpected value for likes count: {likes}")
        return likes

    async def like_component(self, api_key: str, component_id: str) -> bool:
        # if it returns a list with one id, it means the like was successful
        # if it returns an int, it means the like was removed
        if not self.like_webhook_url:
            raise ValueError("LIKE_WEBHOOK_URL is not set")
        headers = {"Authorization": f"Bearer {api_key}"}
        # response = httpx.post(
        #     self.like_webhook_url,
        #     json={"component_id": str(component_id)},
        #     headers=headers,
        # )

        # response.raise_for_status()
        async with httpx.AsyncClient() as client:
            response = await client.post(
                self.like_webhook_url,
                json={"component_id": str(component_id)},
                headers=headers,
            )
            response.raise_for_status()
        if response.status_code == 200:
            result = response.json()

            if isinstance(result, list):
                return True
            elif isinstance(result, int):
                return False
            else:
                raise ValueError(f"Unexpected result: {result}")
        else:
            raise ValueError(f"Unexpected status code: {response.status_code}")

    async def get_list_component_response_model(
        self,
        component_id: Optional[str] = None,
        search: Optional[str] = None,
        private: Optional[bool] = None,
        tags: Optional[List[str]] = None,
        is_component: Optional[bool] = None,
        fields: Optional[List[str]] = None,
        filter_by_user: bool = False,
        liked: bool = False,
        store_api_key: Optional[str] = None,
        sort: Optional[List[str]] = None,
        page: int = 1,
        limit: int = 15,
    ):
        async with user_data_context(api_key=store_api_key, store_service=self):
            filter_conditions: List[Dict[str, Any]] = self.build_filter_conditions(
                component_id=component_id,
                search=search,
                private=private,
                tags=tags,
                is_component=is_component,
                filter_by_user=filter_by_user,
                liked=liked,
                store_api_key=store_api_key,
            )

            result: List[ListComponentResponse] = []
            authorized = False
            metadata: Dict = {}
            comp_count = 0
            try:
                result, metadata = await self.query_components(
                    api_key=store_api_key,
                    page=page,
                    limit=limit,
                    sort=sort,
                    fields=fields,
                    filter_conditions=filter_conditions,
                    use_api_key=liked or filter_by_user,
                )
                if metadata:
                    comp_count = metadata.get("filter_count", 0)
            except HTTPStatusError as exc:
                if exc.response.status_code == 403:
                    raise ForbiddenError("You are not authorized to access this public resource") from exc
                elif exc.response.status_code == 401:
                    raise APIKeyError(
                        "You are not authorized to access this resource. Please check your API key."
                    ) from exc
            except Exception as exc:
                raise ValueError(f"Unexpected error: {exc}") from exc
            try:
                if result and not metadata:
                    if len(result) >= limit:
                        comp_count = await self.count_components(
                            api_key=store_api_key,
                            filter_conditions=filter_conditions,
                            use_api_key=liked or filter_by_user,
                        )
                    else:
                        comp_count = len(result)
                elif not metadata:
                    comp_count = 0
            except HTTPStatusError as exc:
                if exc.response.status_code == 403:
                    raise ForbiddenError("You are not authorized to access this public resource")
                elif exc.response.status_code == 401:
                    raise APIKeyError("You are not authorized to access this resource. Please check your API key.")

            if store_api_key:
                # Now, from the result, we need to get the components
                # the user likes and set the liked_by_user to True
                # if any of the components does not have an id, it means
                # we should not update the components

                if not result or any(component.id is None for component in result):
                    authorized = await self.check_api_key(store_api_key)
                else:
                    try:
                        updated_result = await update_components_with_user_data(
                            result, self, store_api_key, liked=liked
                        )
                        authorized = True
                        result = updated_result
                    except Exception:
                        # If we get an error here, it means the user is not authorized
                        authorized = False
        return ListComponentResponseModel(results=result, authorized=authorized, count=comp_count)



from typing import List, Optional

from langchain.llms.base import BaseLLM
from langchain_community.llms.ollama import Ollama

from langflow import CustomComponent


class OllamaLLM(CustomComponent):
    display_name = "Ollama"
    description = "Local LLM with Ollama."

    def build_config(self) -> dict:
        return {
            "base_url": {
                "display_name": "Base URL",
                "info": "Endpoint of the Ollama API. Defaults to 'http://localhost:11434' if not specified.",
            },
            "model": {
                "display_name": "Model Name",
                "value": "llama2",
                "info": "Refer to https://ollama.ai/library for more models.",
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
                "value": 0.8,
                "info": "Controls the creativity of model responses.",
            },
            "mirostat": {
                "display_name": "Mirostat",
                "options": ["Disabled", "Mirostat", "Mirostat 2.0"],
                "info": "Enable/disable Mirostat sampling for controlling perplexity.",
                "value": "Disabled",
                "advanced": True,
            },
            "mirostat_eta": {
                "display_name": "Mirostat Eta",
                "field_type": "float",
                "info": "Learning rate influencing the algorithm's response to feedback.",
                "advanced": True,
            },
            "mirostat_tau": {
                "display_name": "Mirostat Tau",
                "field_type": "float",
                "info": "Controls balance between coherence and diversity.",
                "advanced": True,
            },
            "num_ctx": {
                "display_name": "Context Window Size",
                "field_type": "int",
                "info": "Size of the context window for generating the next token.",
                "advanced": True,
            },
            "num_gpu": {
                "display_name": "Number of GPUs",
                "field_type": "int",
                "info": "Number of GPUs to use for computation.",
                "advanced": True,
            },
            "num_thread": {
                "display_name": "Number of Threads",
                "field_type": "int",
                "info": "Number of threads to use during computation.",
                "advanced": True,
            },
            "repeat_last_n": {
                "display_name": "Repeat Last N",
                "field_type": "int",
                "info": "Sets how far back the model looks to prevent repetition.",
                "advanced": True,
            },
            "repeat_penalty": {
                "display_name": "Repeat Penalty",
                "field_type": "float",
                "info": "Penalty for repetitions in generated text.",
                "advanced": True,
            },
            "stop": {
                "display_name": "Stop Tokens",
                "info": "List of tokens to signal the model to stop generating text.",
                "advanced": True,
            },
            "tfs_z": {
                "display_name": "TFS Z",
                "field_type": "float",
                "info": "Tail free sampling to reduce impact of less probable tokens.",
                "advanced": True,
            },
            "top_k": {
                "display_name": "Top K",
                "field_type": "int",
                "info": "Limits token selection to top K for reducing nonsense generation.",
                "advanced": True,
            },
            "top_p": {
                "display_name": "Top P",
                "field_type": "int",
                "info": "Works with top-k to control diversity of generated text.",
                "advanced": True,
            },
        }

    def build(
        self,
        base_url: Optional[str],
        model: str,
        temperature: Optional[float],
        mirostat: Optional[str],
        mirostat_eta: Optional[float] = None,
        mirostat_tau: Optional[float] = None,
        num_ctx: Optional[int] = None,
        num_gpu: Optional[int] = None,
        num_thread: Optional[int] = None,
        repeat_last_n: Optional[int] = None,
        repeat_penalty: Optional[float] = None,
        stop: Optional[List[str]] = None,
        tfs_z: Optional[float] = None,
        top_k: Optional[int] = None,
        top_p: Optional[int] = None,
    ) -> BaseLLM:
        if not base_url:
            base_url = "http://localhost:11434"

        # Mapping mirostat settings to their corresponding values
        mirostat_options = {"Mirostat": 1, "Mirostat 2.0": 2}

        # Default to 0 for 'Disabled'
        mirostat_value = mirostat_options.get(mirostat, 0)  # type: ignore

        # Set mirostat_eta and mirostat_tau to None if mirostat is disabled
        if mirostat_value == 0:
            mirostat_eta = None
            mirostat_tau = None

        try:
            llm = Ollama(
                base_url=base_url,
                model=model,
                mirostat=mirostat_value,
                mirostat_eta=mirostat_eta,
                mirostat_tau=mirostat_tau,
                num_ctx=num_ctx,
                num_gpu=num_gpu,
                num_thread=num_thread,
                repeat_last_n=repeat_last_n,
                repeat_penalty=repeat_penalty,
                temperature=temperature,
                stop=stop,
                tfs_z=tfs_z,
                top_k=top_k,
                top_p=top_p,
            )

        except Exception as e:
            raise ValueError("Could not connect to Ollama.") from e

        return llm



from typing import Any, Callable
from celery.result import AsyncResult  # type: ignore
from langflow.services.task.backends.base import TaskBackend
from langflow.worker import celery_app


class CeleryBackend(TaskBackend):
    name = "celery"

    def __init__(self):
        self.celery_app = celery_app

    def launch_task(self, task_func: Callable[..., Any], *args: Any, **kwargs: Any) -> tuple[str, AsyncResult]:
        # I need to type the delay method to make it easier
        from celery import Task  # type: ignore

        if not hasattr(task_func, "delay"):
            raise ValueError(f"Task function {task_func} does not have a delay method")
        task: Task = task_func.delay(*args, **kwargs)
        return task.id, AsyncResult(task.id, app=self.celery_app)

    def get_task(self, task_id: str) -> Any:
        return AsyncResult(task_id, app=self.celery_app)



from typing import Callable, Union

from langflow.template.field.base import TemplateField
from langflow.utils.constants import DIRECT_TYPES
from pydantic import BaseModel, model_serializer


class Template(BaseModel):
    type_name: str
    fields: list[TemplateField]
    field_order: list[str] = []

    def process_fields(
        self,
        format_field_func: Union[Callable, None] = None,
    ):
        if format_field_func:
            for field in self.fields:
                format_field_func(field, self.type_name)

    def sort_fields(self):
        # first sort alphabetically
        # then sort fields so that fields that have .field_type in DIRECT_TYPES are first
        self.fields.sort(key=lambda x: x.name)
        self.fields.sort(key=lambda x: x.field_type in DIRECT_TYPES, reverse=False)

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        result = handler(self)
        for field in self.fields:
            result[field.name] = field.model_dump(by_alias=True, exclude_none=True)
        result["_type"] = result.pop("type_name")
        result.pop("field_order", None)
        return result

    # For backwards compatibility
    def to_dict(self, format_field_func=None):
        self.process_fields(format_field_func)
        self.sort_fields()
        return self.model_dump(by_alias=True, exclude_none=True, exclude={"fields"})

    def add_field(self, field: TemplateField) -> None:
        self.fields.append(field)

    def get_field(self, field_name: str) -> TemplateField:
        """Returns the field with the given name."""
        field = next((field for field in self.fields if field.name == field_name), None)
        if field is None:
            raise ValueError(f"Field {field_name} not found in template {self.type_name}")
        return field

    def update_field(self, field_name: str, field: TemplateField) -> None:
        """Updates the field with the given name."""
        for idx, template_field in enumerate(self.fields):
            if template_field.name == field_name:
                self.fields[idx] = field
                return
        raise ValueError(f"Field {field_name} not found in template {self.type_name}")

    def upsert_field(self, field_name: str, field: TemplateField) -> None:
        """Updates the field with the given name or adds it if it doesn't exist."""
        try:
            self.update_field(field_name, field)
        except ValueError:
            self.add_field(field)



from langflow import CustomComponent
from langchain_community.tools.json.tool import JsonSpec
from langchain_community.agent_toolkits.json.toolkit import JsonToolkit


class JsonToolkitComponent(CustomComponent):
    display_name = "JsonToolkit"
    description = "Toolkit for interacting with a JSON spec."

    def build_config(self):
        return {
            "spec": {"display_name": "Spec", "type": JsonSpec},
        }

    def build(self, spec: JsonSpec) -> JsonToolkit:
        return JsonToolkit(spec=spec)



from typing import Callable, List, Optional, Union

from langchain.agents import AgentExecutor, AgentType, initialize_agent, types
from langflow import CustomComponent
from langflow.field_typing import BaseChatMemory, BaseLanguageModel, Tool


class AgentInitializerComponent(CustomComponent):
    display_name: str = "Agent Initializer"
    description: str = "Initialize a Langchain Agent."
    documentation: str = "https://python.langchain.com/docs/modules/agents/agent_types/"

    def build_config(self):
        agents = list(types.AGENT_TO_CLASS.keys())
        # field_type and required are optional
        return {
            "agent": {"options": agents, "value": agents[0], "display_name": "Agent Type"},
            "max_iterations": {"display_name": "Max Iterations", "value": 10},
            "memory": {"display_name": "Memory"},
            "tools": {"display_name": "Tools"},
            "llm": {"display_name": "Language Model"},
            "code": {"advanced": True},
        }

    def build(
        self,
        agent: str,
        llm: BaseLanguageModel,
        tools: List[Tool],
        max_iterations: int,
        memory: Optional[BaseChatMemory] = None,
    ) -> Union[AgentExecutor, Callable]:
        agent = AgentType(agent)
        if memory:
            return initialize_agent(
                tools=tools,
                llm=llm,
                agent=agent,
                memory=memory,
                return_intermediate_steps=True,
                handle_parsing_errors=True,
                max_iterations=max_iterations,
            )
        return initialize_agent(
            tools=tools,
            llm=llm,
            agent=agent,
            return_intermediate_steps=True,
            handle_parsing_errors=True,
            max_iterations=max_iterations,
        )



"""Add unique constraints

Revision ID: 006b3990db50
Revises: 1ef9c4f3765d
Create Date: 2023-12-13 18:55:52.587360

"""

from typing import Sequence, Union

from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision: str = "006b3990db50"
down_revision: Union[str, None] = "1ef9c4f3765d"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    api_key_constraints = inspector.get_unique_constraints("apikey")
    flow_constraints = inspector.get_unique_constraints("flow")
    user_constraints = inspector.get_unique_constraints("user")
    try:
        if not any(constraint['column_names'] == ['id'] for constraint in api_key_constraints):
            with op.batch_alter_table("apikey", schema=None) as batch_op:

                batch_op.create_unique_constraint("uq_apikey_id", ["id"])
        if not any(constraint["column_names"] == ["id"] for constraint in flow_constraints):
            with op.batch_alter_table("flow", schema=None) as batch_op:
                batch_op.create_unique_constraint("uq_flow_id", ["id"])
        if not any(constraint["column_names"] == ["id"] for constraint in user_constraints):
            with op.batch_alter_table("user", schema=None) as batch_op:
                batch_op.create_unique_constraint("uq_user_id", ["id"])
    except Exception as e:
        print(e)
        pass

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    api_key_constraints = inspector.get_unique_constraints("apikey")
    flow_constraints = inspector.get_unique_constraints("flow")
    user_constraints = inspector.get_unique_constraints("user")
    try:
        if any(
            constraint["name"] == "uq_apikey_id" for constraint in api_key_constraints
        ):
            with op.batch_alter_table("user", schema=None) as batch_op:
                batch_op.drop_constraint("uq_user_id", type_="unique")
        if any(constraint["name"] == "uq_flow_id" for constraint in flow_constraints):
            with op.batch_alter_table("flow", schema=None) as batch_op:
                batch_op.drop_constraint("uq_flow_id", type_="unique")
        if any(constraint["name"] == "uq_user_id" for constraint in user_constraints):

            with op.batch_alter_table("apikey", schema=None) as batch_op:
                batch_op.drop_constraint("uq_apikey_id", type_="unique")
    except Exception as e:
        print(e)
        pass
    # ### end Alembic commands ###



from typing import Optional

import requests
from langchain_core.documents import Document
from langflow import CustomComponent
from langflow.services.database.models.base import orjson_dumps


class GetRequest(CustomComponent):
    display_name: str = "GET Request"
    description: str = "Make a GET request to the given URL."
    output_types: list[str] = ["Document"]
    documentation: str = "https://docs.langflow.org/components/utilities#get-request"
    beta: bool = True
    field_config = {
        "url": {
            "display_name": "URL",
            "info": "The URL to make the request to",
            "is_list": True,
        },
        "headers": {
            "display_name": "Headers",
            "info": "The headers to send with the request.",
        },
        "code": {"show": False},
        "timeout": {
            "display_name": "Timeout",
            "field_type": "int",
            "info": "The timeout to use for the request.",
            "value": 5,
        },
    }

    def get_document(self, session: requests.Session, url: str, headers: Optional[dict], timeout: int) -> Document:
        try:
            response = session.get(url, headers=headers, timeout=int(timeout))
            try:
                response_json = response.json()
                result = orjson_dumps(response_json, indent_2=False)
            except Exception:
                result = response.text
            self.repr_value = result
            return Document(
                page_content=result,
                metadata={
                    "source": url,
                    "headers": headers,
                    "status_code": response.status_code,
                },
            )
        except requests.Timeout:
            return Document(
                page_content="Request Timed Out",
                metadata={"source": url, "headers": headers, "status_code": 408},
            )
        except Exception as exc:
            return Document(
                page_content=str(exc),
                metadata={"source": url, "headers": headers, "status_code": 500},
            )

    def build(
        self,
        url: str,
        headers: Optional[dict] = None,
        timeout: int = 5,
    ) -> list[Document]:
        if headers is None:
            headers = {}
        urls = url if isinstance(url, list) else [url]
        with requests.Session() as session:
            documents = [self.get_document(session, u, headers, timeout) for u in urls]
            self.repr_value = documents
        return documents



import asyncio
import uuid
from collections import defaultdict
from typing import Any, Dict, List

import orjson
from fastapi import WebSocket, status
from loguru import logger
from starlette.websockets import WebSocketState

from langflow.api.v1.schemas import ChatMessage, ChatResponse, FileResponse
from langflow.interface.utils import pil_to_base64
from langflow.services import ServiceType, service_manager
from langflow.services.base import Service
from langflow.services.chat.cache import Subject
from langflow.services.chat.utils import process_graph

from .cache import cache_service


class ChatHistory(Subject):
    def __init__(self):
        super().__init__()
        self.history: Dict[str, List[ChatMessage]] = defaultdict(list)

    def add_message(self, client_id: str, message: ChatMessage):
        """Add a message to the chat history."""

        self.history[client_id].append(message)

        if not isinstance(message, FileResponse):
            self.notify()

    def get_history(self, client_id: str, filter_messages=True) -> List[ChatMessage]:
        """Get the chat history for a client."""
        if history := self.history.get(client_id, []):
            if filter_messages:
                return [msg for msg in history if msg.type not in ["start", "stream"]]
            return history
        else:
            return []

    def empty_history(self, client_id: str):
        """Empty the chat history for a client."""
        self.history[client_id] = []


class ChatService(Service):
    name = "chat_service"

    def __init__(self):
        self.active_connections: Dict[str, WebSocket] = {}
        self.connection_ids: Dict[str, str] = {}
        self.chat_history = ChatHistory()
        self.chat_cache = cache_service
        self.chat_cache.attach(self.update)
        self.cache_service = service_manager.get(ServiceType.CACHE_SERVICE)

    def on_chat_history_update(self):
        """Send the last chat message to the client."""
        client_id = self.chat_cache.current_client_id
        if client_id in self.active_connections:
            chat_response = self.chat_history.get_history(client_id, filter_messages=False)[-1]
            if chat_response.is_bot:
                # Process FileResponse
                if isinstance(chat_response, FileResponse):
                    # If data_type is pandas, convert to csv
                    if chat_response.data_type == "pandas":
                        chat_response.data = chat_response.data.to_csv()
                    elif chat_response.data_type == "image":
                        # Base64 encode the image
                        chat_response.data = pil_to_base64(chat_response.data)
                # get event loop
                loop = asyncio.get_event_loop()

                coroutine = self.send_json(client_id, chat_response)
                asyncio.run_coroutine_threadsafe(coroutine, loop)

    def update(self):
        if self.chat_cache.current_client_id in self.active_connections:
            self.last_cached_object_dict = self.chat_cache.get_last()
            # Add a new ChatResponse with the data
            chat_response = FileResponse(
                message=None,
                type="file",
                data=self.last_cached_object_dict["obj"],
                data_type=self.last_cached_object_dict["type"],
            )

            self.chat_history.add_message(self.chat_cache.current_client_id, chat_response)

    async def connect(self, client_id: str, websocket: WebSocket):
        self.active_connections[client_id] = websocket
        # This is to avoid having multiple clients with the same id
        #! Temporary solution
        self.connection_ids[client_id] = f"{client_id}-{uuid.uuid4()}"

    def disconnect(self, client_id: str):
        self.active_connections.pop(client_id, None)
        self.connection_ids.pop(client_id, None)

    async def send_message(self, client_id: str, message: str):
        websocket = self.active_connections[client_id]
        await websocket.send_text(message)

    async def send_json(self, client_id: str, message: ChatMessage):
        websocket = self.active_connections[client_id]
        await websocket.send_json(message.model_dump())

    async def close_connection(self, client_id: str, code: int, reason: str):
        if websocket := self.active_connections[client_id]:
            try:
                await websocket.close(code=code, reason=reason)
                self.disconnect(client_id)
            except RuntimeError as exc:
                # This is to catch the following error:
                #  Unexpected ASGI message 'websocket.close', after sending 'websocket.close'
                if "after sending" in str(exc):
                    logger.error(f"Error closing connection: {exc}")

    async def process_message(self, client_id: str, payload: Dict, build_result: Any):
        # Process the graph data and chat message
        chat_inputs = payload.pop("inputs", {})
        chatkey = payload.pop("chatKey", None)
        chat_inputs = ChatMessage(message=chat_inputs, chatKey=chatkey)
        self.chat_history.add_message(client_id, chat_inputs)

        # graph_data = payload
        start_resp = ChatResponse(message=None, type="start", intermediate_steps="")
        await self.send_json(client_id, start_resp)

        # is_first_message = len(self.chat_history.get_history(client_id=client_id)) <= 1
        # Generate result and thought
        try:
            logger.debug("Generating result and thought")

            result, intermediate_steps, raw_output = await process_graph(
                build_result=build_result,
                chat_inputs=chat_inputs,
                client_id=client_id,
                session_id=self.connection_ids[client_id],
            )
            self.set_cache(client_id, build_result)
        except Exception as e:
            # Log stack trace
            logger.exception(e)
            self.chat_history.empty_history(client_id)
            raise e
        # Send a response back to the frontend, if needed
        intermediate_steps = intermediate_steps or ""
        history = self.chat_history.get_history(client_id, filter_messages=False)
        file_responses = []
        if history:
            # Iterate backwards through the history
            for msg in reversed(history):
                if isinstance(msg, FileResponse):
                    if msg.data_type == "image":
                        # Base64 encode the image
                        if isinstance(msg.data, str):
                            continue
                        msg.data = pil_to_base64(msg.data)
                    file_responses.append(msg)
                if msg.type == "start":
                    break

        response = ChatResponse(
            message=result,
            intermediate_steps=intermediate_steps.strip(),
            type="end",
            files=file_responses,
        )
        await self.send_json(client_id, response)
        self.chat_history.add_message(client_id, response)

    def set_cache(self, client_id: str, langchain_object: Any) -> bool:
        """
        Set the cache for a client.
        """
        # client_id is the flow id but that already exists in the cache
        # so we need to change it to something else

        result_dict = {
            "result": langchain_object,
            "type": type(langchain_object),
        }
        self.cache_service.upsert(client_id, result_dict)
        return client_id in self.cache_service

    async def handle_websocket(self, client_id: str, websocket: WebSocket):
        await self.connect(client_id, websocket)

        try:
            chat_history = self.chat_history.get_history(client_id)
            # iterate and make BaseModel into dict
            chat_history = [chat.model_dump() for chat in chat_history]
            await websocket.send_json(chat_history)

            while True:
                json_payload = await websocket.receive_json()
                if isinstance(json_payload, str):
                    payload = orjson.loads(json_payload)
                elif isinstance(json_payload, dict):
                    payload = json_payload
                if "clear_history" in payload and payload["clear_history"]:
                    self.chat_history.history[client_id] = []
                    continue

                with self.chat_cache.set_client_id(client_id):
                    if build_result := self.cache_service.get(client_id).get("result"):
                        await self.process_message(client_id, payload, build_result)

                    else:
                        raise RuntimeError(f"Could not find a build result for client_id {client_id}")
        except Exception as exc:
            # Handle any exceptions that might occur
            logger.exception(f"Error handling websocket: {exc}")
            if websocket.client_state == WebSocketState.CONNECTED:
                await self.close_connection(
                    client_id=client_id,
                    code=status.WS_1011_INTERNAL_ERROR,
                    reason=str(exc),
                )
            elif websocket.client_state == WebSocketState.DISCONNECTED:
                self.disconnect(client_id)

        finally:
            try:
                # first check if the connection is still open
                if websocket.client_state == WebSocketState.CONNECTED:
                    await self.close_connection(
                        client_id=client_id,
                        code=status.WS_1000_NORMAL_CLOSURE,
                        reason="Client disconnected",
                    )
            except Exception as exc:
                logger.error(f"Error closing connection: {exc}")
            self.disconnect(client_id)


def dict_to_markdown_table(my_dict):
    markdown_table = "| Key | Value |\n|---|---|\n"
    for key, value in my_dict.items():
        markdown_table += f"| {key} | {value} |\n"
    return markdown_table


def list_of_dicts_to_markdown_table(dict_list):
    if not dict_list:
        return "No data provided."

    # Extract headers from the keys of the first dictionary
    headers = dict_list[0].keys()
    markdown_table = "| " + " | ".join(headers) + " |\n"
    markdown_table += "| " + " | ".join("---" for _ in headers) + " |\n"

    for row_dict in dict_list:
        row = [str(row_dict.get(header, "")) for header in headers]
        markdown_table += "| " + " | ".join(row) + " |\n"

    return markdown_table



"""Add unique constraints

Revision ID: b2fa308044b5
Revises: 0b8757876a7c
Create Date: 2024-01-26 13:31:14.797548

"""

from typing import Sequence, Union

import sqlalchemy as sa
import sqlmodel
from alembic import op
from sqlalchemy.engine.reflection import Inspector

# revision identifiers, used by Alembic.
revision: str = "b2fa308044b5"
down_revision: Union[str, None] = "0b8757876a7c"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    tables = inspector.get_table_names()
    # ### commands auto generated by Alembic - please adjust! ###
    try:
        if "flowstyle" in tables:
            op.drop_table("flowstyle")
        with op.batch_alter_table("flow", schema=None) as batch_op:
            flow_columns = [column["name"] for column in inspector.get_columns("flow")]
            if "is_component" not in flow_columns:
                batch_op.add_column(
                    sa.Column("is_component", sa.Boolean(), nullable=True)
                )
            if "updated_at" not in flow_columns:
                batch_op.add_column(
                    sa.Column("updated_at", sa.DateTime(), nullable=True)
                )
            if "folder" not in flow_columns:
                batch_op.add_column(
                    sa.Column(
                        "folder", sqlmodel.sql.sqltypes.AutoString(), nullable=True
                    )
                )
            if "user_id" not in flow_columns:
                batch_op.add_column(
                    sa.Column("user_id", sqlmodel.sql.sqltypes.GUID(), nullable=True)
                )
            indices = inspector.get_indexes("flow")
            indices_names = [index["name"] for index in indices]
            if "ix_flow_user_id" not in indices_names:
                batch_op.create_index(
                    batch_op.f("ix_flow_user_id"), ["user_id"], unique=False
                )
            if "fk_flow_user_id_user" not in indices_names:
                batch_op.create_foreign_key(
                    "fk_flow_user_id_user", "user", ["user_id"], ["id"]
                )

    except Exception:
        pass
    # ### end Alembic commands ###


def downgrade() -> None:
    conn = op.get_bind()
    inspector = Inspector.from_engine(conn)  # type: ignore
    try:
        # Re-create the dropped table 'flowstyle' if it was previously dropped in upgrade
        if "flowstyle" not in inspector.get_table_names():
            op.create_table(
                "flowstyle",
                sa.Column("color", sa.String(), nullable=False),
                sa.Column("emoji", sa.String(), nullable=False),
                sa.Column("flow_id", sqlmodel.sql.sqltypes.GUID(), nullable=True),
                sa.Column("id", sqlmodel.sql.sqltypes.GUID(), nullable=False),
                sa.ForeignKeyConstraint(["flow_id"], ["flow.id"]),
                sa.PrimaryKeyConstraint("id"),
                sa.UniqueConstraint("id"),
            )

        with op.batch_alter_table("flow", schema=None) as batch_op:
            # Check and remove newly added columns and constraints in upgrade
            flow_columns = [column["name"] for column in inspector.get_columns("flow")]
            if "user_id" in flow_columns:
                batch_op.drop_column("user_id")
            if "folder" in flow_columns:
                batch_op.drop_column("folder")
            if "updated_at" in flow_columns:
                batch_op.drop_column("updated_at")
            if "is_component" in flow_columns:
                batch_op.drop_column("is_component")

            indices = inspector.get_indexes("flow")
            indices_names = [index["name"] for index in indices]
            if "ix_flow_user_id" in indices_names:
                batch_op.drop_index("ix_flow_user_id")
            # Assuming fk_flow_user_id_user is a foreign key constraint's name, not an index
            constraints = inspector.get_foreign_keys("flow")
            constraint_names = [constraint["name"] for constraint in constraints]
            if "fk_flow_user_id_user" in constraint_names:
                batch_op.drop_constraint("fk_flow_user_id_user", type_="foreignkey")

    except Exception as e:
        # It's generally a good idea to log the exception or handle it in a way other than a bare pass
        print(f"Error during downgrade: {e}")



from langchain.chains.llm import LLMChain
from langchain.agents import AgentExecutor, ZeroShotAgent
from langchain.agents.agent_toolkits.json.prompt import JSON_PREFIX, JSON_SUFFIX
from langchain.agents.agent_toolkits.json.toolkit import JsonToolkit
from langchain.agents.mrkl.prompt import FORMAT_INSTRUCTIONS
from langchain.base_language import BaseLanguageModel


class MalfoyAgent(AgentExecutor):
    """Json agent"""

    prefix = "Malfoy: "

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(cls, toolkit: JsonToolkit, llm: BaseLanguageModel):
        tools = toolkit.get_tools()
        tool_names = {tool.name for tool in tools}
        prompt = ZeroShotAgent.create_prompt(
            tools,
            prefix=JSON_PREFIX,
            suffix=JSON_SUFFIX,
            format_instructions=FORMAT_INSTRUCTIONS,
            input_variables=None,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
        )
        agent = ZeroShotAgent(llm_chain=llm_chain, allowed_tools=tool_names)  # type: ignore
        return cls.from_agent_and_tools(agent=agent, tools=tools, verbose=True)

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


PREBUILT_AGENTS = {
    "MalfoyAgent": MalfoyAgent,
}



from typing import Optional

from langflow import CustomComponent
from langchain.embeddings.base import Embeddings
from langchain_community.embeddings import OllamaEmbeddings


class OllamaEmbeddingsComponent(CustomComponent):
    """
    A custom component for implementing an Embeddings Model using Ollama.
    """

    display_name: str = "Ollama Embeddings"
    description: str = "Embeddings model from Ollama."
    documentation = "https://python.langchain.com/docs/integrations/text_embedding/ollama"
    beta = True

    def build_config(self):
        return {
            "model": {
                "display_name": "Ollama Model",
            },
            "base_url": {"display_name": "Ollama Base URL"},
            "temperature": {"display_name": "Model Temperature"},
            "code": {"show": False},
        }

    def build(
        self,
        model: str = "llama2",
        base_url: str = "http://localhost:11434",
        temperature: Optional[float] = None,
    ) -> Embeddings:
        try:
            output = OllamaEmbeddings(model=model, base_url=base_url, temperature=temperature)  # type: ignore
        except Exception as e:
            raise ValueError("Could not connect to Ollama API.") from e
        return output



from pathlib import Path
from langflow.services.settings.service import SettingsService
from langflow.services.factory import ServiceFactory


class SettingsServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(SettingsService)

    def create(self):
        # Here you would have logic to create and configure a SettingsService
        langflow_dir = Path(__file__).parent.parent.parent
        return SettingsService.load_settings_from_yaml(str(langflow_dir / "config.yaml"))



import asyncio
from typing import Any, Dict, List, Optional
from uuid import UUID

from langchain.schema import AgentAction, AgentFinish
from langchain_core.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler
from langflow.api.v1.schemas import ChatResponse, PromptResponse
from langflow.services.deps import get_chat_service
from langflow.utils.util import remove_ansi_escape_codes
from loguru import logger


# https://github.com/hwchase17/chat-langchain/blob/master/callback.py
class AsyncStreamingLLMCallbackHandler(AsyncCallbackHandler):
    """Callback handler for streaming LLM responses."""

    @property
    def ignore_chain(self) -> bool:
        """Whether to ignore chain callbacks."""
        return False

    def __init__(self, client_id: str):
        self.chat_service = get_chat_service()
        self.client_id = client_id
        self.websocket = self.chat_service.active_connections[self.client_id]

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        resp = ChatResponse(message=token, type="stream", intermediate_steps="")
        await self.websocket.send_json(resp.model_dump())

    async def on_tool_start(self, serialized: Dict[str, Any], input_str: str, **kwargs: Any) -> Any:
        """Run when tool starts running."""
        resp = ChatResponse(
            message="",
            type="stream",
            intermediate_steps=f"Tool input: {input_str}",
        )
        await self.websocket.send_json(resp.model_dump())

    async def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        """Run when tool ends running."""
        observation_prefix = kwargs.get("observation_prefix", "Tool output: ")
        split_output = output.split()
        first_word = split_output[0]
        rest_of_output = split_output[1:]
        # Create a formatted message.
        intermediate_steps = f"{observation_prefix}{first_word}"

        # Create a ChatResponse instance.
        resp = ChatResponse(
            message="",
            type="stream",
            intermediate_steps=intermediate_steps,
        )
        rest_of_resps = [
            ChatResponse(
                message="",
                type="stream",
                intermediate_steps=f"{word}",
            )
            for word in rest_of_output
        ]
        resps = [resp] + rest_of_resps
        # Try to send the response, handle potential errors.

        try:
            # This is to emulate the stream of tokens
            for resp in resps:
                await self.websocket.send_json(resp.model_dump())
        except Exception as exc:
            logger.error(f"Error sending response: {exc}")

    async def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> None:
        """Run when tool errors."""

    async def on_text(self, text: str, **kwargs: Any) -> Any:
        """Run on arbitrary text."""
        # This runs when first sending the prompt
        # to the LLM, adding it will send the final prompt
        # to the frontend
        if "Prompt after formatting" in text:
            text = text.replace("Prompt after formatting:\n", "")
            text = remove_ansi_escape_codes(text)
            resp = PromptResponse(
                prompt=text,
            )
            await self.websocket.send_json(resp.model_dump())
            self.chat_service.chat_history.add_message(self.client_id, resp)

    async def on_agent_action(self, action: AgentAction, **kwargs: Any):
        log = f"Thought: {action.log}"
        # if there are line breaks, split them and send them
        # as separate messages
        if "\n" in log:
            logs = log.split("\n")
            for log in logs:
                resp = ChatResponse(message="", type="stream", intermediate_steps=log)
                await self.websocket.send_json(resp.model_dump())
        else:
            resp = ChatResponse(message="", type="stream", intermediate_steps=log)
            await self.websocket.send_json(resp.model_dump())

    async def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> Any:
        """Run on agent end."""
        resp = ChatResponse(
            message="",
            type="stream",
            intermediate_steps=finish.log,
        )
        await self.websocket.send_json(resp.model_dump())


class StreamingLLMCallbackHandler(BaseCallbackHandler):
    """Callback handler for streaming LLM responses."""

    def __init__(self, client_id: str):
        self.chat_service = get_chat_service()
        self.client_id = client_id
        self.websocket = self.chat_service.active_connections[self.client_id]

    def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        resp = ChatResponse(message=token, type="stream", intermediate_steps="")

        loop = asyncio.get_event_loop()
        coroutine = self.websocket.send_json(resp.model_dump())
        asyncio.run_coroutine_threadsafe(coroutine, loop)



from typing import Optional

from langchain.prompts import PromptTemplate
from pydantic import BaseModel, field_validator, model_serializer

from langflow.interface.utils import extract_input_variables_from_prompt
from langflow.template.frontend_node.base import FrontendNode


class CacheResponse(BaseModel):
    data: dict


class Code(BaseModel):
    code: str


class FrontendNodeRequest(FrontendNode):
    template: dict  # type: ignore

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        # Override the default serialization method in FrontendNode
        # because we don't need the name in the response (i.e. {name: {}})
        return handler(self)


class ValidatePromptRequest(BaseModel):
    name: str
    template: str
    # optional for tweak call
    frontend_node: Optional[FrontendNodeRequest] = None


# Build ValidationResponse class for {"imports": {"errors": []}, "function": {"errors": []}}
class CodeValidationResponse(BaseModel):
    imports: dict
    function: dict

    @field_validator("imports")
    @classmethod
    def validate_imports(cls, v):
        return v or {"errors": []}

    @field_validator("function")
    @classmethod
    def validate_function(cls, v):
        return v or {"errors": []}


class PromptValidationResponse(BaseModel):
    input_variables: list
    # object return for tweak call
    frontend_node: Optional[FrontendNodeRequest] = None


INVALID_CHARACTERS = {
    " ",
    ",",
    ".",
    ":",
    ";",
    "!",
    "?",
    "/",
    "\\",
    "(",
    ")",
    "[",
    "]",
    "{",
    "}",
}

INVALID_NAMES = {
    "input_variables",
    "output_parser",
    "partial_variables",
    "template",
    "template_format",
    "validate_template",
}


def validate_prompt(template: str):
    input_variables = extract_input_variables_from_prompt(template)

    # Check if there are invalid characters in the input_variables
    input_variables = check_input_variables(input_variables)
    if any(var in INVALID_NAMES for var in input_variables):
        raise ValueError(f"Invalid input variables. None of the variables can be named {', '.join(input_variables)}. ")

    try:
        PromptTemplate(template=template, input_variables=input_variables)
    except Exception as exc:
        raise ValueError(str(exc)) from exc

    return input_variables


def check_input_variables(input_variables: list):
    invalid_chars = []
    fixed_variables = []
    wrong_variables = []
    empty_variables = []
    for variable in input_variables:
        new_var = variable

        # if variable is empty, then we should add that to the wrong variables
        if not variable:
            empty_variables.append(variable)
            continue

        # if variable starts with a number we should add that to the invalid chars
        # and wrong variables
        if variable[0].isdigit():
            invalid_chars.append(variable[0])
            new_var = new_var.replace(variable[0], "")
            wrong_variables.append(variable)
        else:
            for char in INVALID_CHARACTERS:
                if char in variable:
                    invalid_chars.append(char)
                    new_var = new_var.replace(char, "")
                    wrong_variables.append(variable)
        fixed_variables.append(new_var)
    # If any of the input_variables is not in the fixed_variables, then it means that
    # there are invalid characters in the input_variables

    if any(var not in fixed_variables for var in input_variables):
        error_message = build_error_message(
            input_variables,
            invalid_chars,
            wrong_variables,
            fixed_variables,
            empty_variables,
        )
        raise ValueError(error_message)
    return input_variables


def build_error_message(input_variables, invalid_chars, wrong_variables, fixed_variables, empty_variables):
    input_variables_str = ", ".join([f"'{var}'" for var in input_variables])
    error_string = f"Invalid input variables: {input_variables_str}. "

    if wrong_variables and invalid_chars:
        # fix the wrong variables replacing invalid chars and find them in the fixed variables
        error_string_vars = "You can fix them by replacing the invalid characters: "
        wvars = wrong_variables.copy()
        for i, wrong_var in enumerate(wvars):
            for char in invalid_chars:
                wrong_var = wrong_var.replace(char, "")
            if wrong_var in fixed_variables:
                error_string_vars += f"'{wrong_variables[i]}' -> '{wrong_var}'"
        error_string += error_string_vars
    elif empty_variables:
        error_string += f" There are {len(empty_variables)} empty variable{'s' if len(empty_variables) > 1 else ''}."
    elif len(set(fixed_variables)) != len(fixed_variables):
        error_string += "There are duplicate variables."
    return error_string



from typing import Callable, Union

from langchain.agents.agent_toolkits.vectorstore.toolkit import VectorStoreInfo
from langchain_community.vectorstores import VectorStore

from langflow import CustomComponent


class VectorStoreInfoComponent(CustomComponent):
    display_name = "VectorStoreInfo"
    description = "Information about a VectorStore"

    def build_config(self):
        return {
            "vectorstore": {"display_name": "VectorStore"},
            "description": {"display_name": "Description", "multiline": True},
            "name": {"display_name": "Name"},
        }

    def build(
        self,
        vectorstore: VectorStore,
        description: str,
        name: str,
    ) -> Union[VectorStoreInfo, Callable]:
        return VectorStoreInfo(vectorstore=vectorstore, description=description, name=name)



from typing import Optional, Union

from langchain.schema import BaseRetriever
from langchain_community.vectorstores import VectorStore
from langchain_community.vectorstores.qdrant import Qdrant
from langflow import CustomComponent
from langflow.field_typing import Document, Embeddings, NestedDict


class QdrantComponent(CustomComponent):
    display_name = "Qdrant"
    description = "Construct Qdrant wrapper from a list of texts."

    def build_config(self):
        return {
            "documents": {"display_name": "Documents"},
            "embedding": {"display_name": "Embedding"},
            "api_key": {"display_name": "API Key", "password": True, "advanced": True},
            "collection_name": {"display_name": "Collection Name"},
            "content_payload_key": {"display_name": "Content Payload Key", "advanced": True},
            "distance_func": {"display_name": "Distance Function", "advanced": True},
            "grpc_port": {"display_name": "gRPC Port", "advanced": True},
            "host": {"display_name": "Host", "advanced": True},
            "https": {"display_name": "HTTPS", "advanced": True},
            "location": {"display_name": "Location", "advanced": True},
            "metadata_payload_key": {"display_name": "Metadata Payload Key", "advanced": True},
            "path": {"display_name": "Path", "advanced": True},
            "port": {"display_name": "Port", "advanced": True},
            "prefer_grpc": {"display_name": "Prefer gRPC", "advanced": True},
            "prefix": {"display_name": "Prefix", "advanced": True},
            "search_kwargs": {"display_name": "Search Kwargs", "advanced": True},
            "timeout": {"display_name": "Timeout", "advanced": True},
            "url": {"display_name": "URL", "advanced": True},
        }

    def build(
        self,
        embedding: Embeddings,
        collection_name: str,
        documents: Optional[Document] = None,
        api_key: Optional[str] = None,
        content_payload_key: str = "page_content",
        distance_func: str = "Cosine",
        grpc_port: int = 6334,
        https: bool = False,
        host: Optional[str] = None,
        location: Optional[str] = None,
        metadata_payload_key: str = "metadata",
        path: Optional[str] = None,
        port: Optional[int] = 6333,
        prefer_grpc: bool = False,
        prefix: Optional[str] = None,
        search_kwargs: Optional[NestedDict] = None,
        timeout: Optional[int] = None,
        url: Optional[str] = None,
    ) -> Union[VectorStore, Qdrant, BaseRetriever]:
        if documents is None:
            from qdrant_client import QdrantClient

            client = QdrantClient(
                location=location,
                url=host,
                port=port,
                grpc_port=grpc_port,
                https=https,
                prefix=prefix,
                timeout=timeout,
                prefer_grpc=prefer_grpc,
                metadata_payload_key=metadata_payload_key,
                content_payload_key=content_payload_key,
                api_key=api_key,
                collection_name=collection_name,
                host=host,
                path=path,
            )
            vs = Qdrant(
                client=client,
                collection_name=collection_name,
                embeddings=embedding,
            )
            return vs
        else:
            vs = Qdrant.from_documents(
                documents=documents,  # type: ignore
                embedding=embedding,
                api_key=api_key,
                collection_name=collection_name,
                content_payload_key=content_payload_key,
                distance_func=distance_func,
                grpc_port=grpc_port,
                host=host,
                https=https,
                location=location,
                metadata_payload_key=metadata_payload_key,
                path=path,
                port=port,
                prefer_grpc=prefer_grpc,
                prefix=prefix,
                search_kwargs=search_kwargs,
                timeout=timeout,
                url=url,
            )
        return vs



from typing import Optional

from langchain_community.embeddings.cohere import CohereEmbeddings

from langflow import CustomComponent


class CohereEmbeddingsComponent(CustomComponent):
    display_name = "CohereEmbeddings"
    description = "Cohere embedding models."

    def build_config(self):
        return {
            "cohere_api_key": {"display_name": "Cohere API Key", "password": True},
            "model": {"display_name": "Model", "default": "embed-english-v2.0", "advanced": True},
            "truncate": {"display_name": "Truncate", "advanced": True},
            "max_retries": {"display_name": "Max Retries", "advanced": True},
            "user_agent": {"display_name": "User Agent", "advanced": True},
        }

    def build(
        self,
        request_timeout: Optional[float] = None,
        cohere_api_key: str = "",
        max_retries: int = 3,
        model: str = "embed-english-v2.0",
        truncate: Optional[str] = None,
        user_agent: str = "langchain",
    ) -> CohereEmbeddings:
        return CohereEmbeddings(  # type: ignore
            max_retries=max_retries,
            user_agent=user_agent,
            request_timeout=request_timeout,
            cohere_api_key=cohere_api_key,
            model=model,
            truncate=truncate,
        )



import inspect
import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Type

import orjson
from langchain.agents import agent as agent_module
from langchain.agents.agent import AgentExecutor
from langchain.agents.agent_toolkits.base import BaseToolkit
from langchain.agents.tools import BaseTool
from langchain.chains.base import Chain
from langchain.document_loaders.base import BaseLoader
from langchain_community.vectorstores import VectorStore
from langchain_core.documents import Document
from loguru import logger
from pydantic import ValidationError

from langflow.interface.custom_lists import CUSTOM_NODES
from langflow.interface.importing.utils import eval_custom_component_code, get_function, import_by_type
from langflow.interface.initialize.llm import initialize_vertexai
from langflow.interface.initialize.utils import handle_format_kwargs, handle_node_type, handle_partial_variables
from langflow.interface.initialize.vector_store import vecstore_initializer
from langflow.interface.output_parsers.base import output_parser_creator
from langflow.interface.retrievers.base import retriever_creator
from langflow.interface.toolkits.base import toolkits_creator
from langflow.interface.utils import load_file_into_dict
from langflow.interface.wrappers.base import wrapper_creator
from langflow.utils import validate

if TYPE_CHECKING:
    from langflow import CustomComponent


def build_vertex_in_params(params: Dict) -> Dict:
    from langflow.graph.vertex.base import Vertex

    # If any of the values in params is a Vertex, we will build it
    return {key: value.build() if isinstance(value, Vertex) else value for key, value in params.items()}


async def instantiate_class(node_type: str, base_type: str, params: Dict, user_id=None) -> Any:
    """Instantiate class from module type and key, and params"""
    params = convert_params_to_sets(params)
    params = convert_kwargs(params)

    if node_type in CUSTOM_NODES:
        if custom_node := CUSTOM_NODES.get(node_type):
            if hasattr(custom_node, "initialize"):
                return custom_node.initialize(**params)
            return custom_node(**params)
    logger.debug(f"Instantiating {node_type} of type {base_type}")
    class_object = import_by_type(_type=base_type, name=node_type)
    return await instantiate_based_on_type(class_object, base_type, node_type, params, user_id=user_id)


def convert_params_to_sets(params):
    """Convert certain params to sets"""
    if "allowed_special" in params:
        params["allowed_special"] = set(params["allowed_special"])
    if "disallowed_special" in params:
        params["disallowed_special"] = set(params["disallowed_special"])
    return params


def convert_kwargs(params):
    # if *kwargs are passed as a string, convert to dict
    # first find any key that has kwargs or config in it
    kwargs_keys = [key for key in params.keys() if "kwargs" in key or "config" in key]
    for key in kwargs_keys:
        if isinstance(params[key], str):
            try:
                params[key] = orjson.loads(params[key])
            except json.JSONDecodeError:
                # if the string is not a valid json string, we will
                # remove the key from the params
                params.pop(key, None)
    return params


async def instantiate_based_on_type(class_object, base_type, node_type, params, user_id):
    if base_type == "agents":
        return instantiate_agent(node_type, class_object, params)
    elif base_type == "prompts":
        return instantiate_prompt(node_type, class_object, params)
    elif base_type == "tools":
        tool = instantiate_tool(node_type, class_object, params)
        if hasattr(tool, "name") and isinstance(tool, BaseTool):
            # tool name shouldn't contain spaces
            tool.name = tool.name.replace(" ", "_")
        return tool
    elif base_type == "toolkits":
        return instantiate_toolkit(node_type, class_object, params)
    elif base_type == "embeddings":
        return instantiate_embedding(node_type, class_object, params)
    elif base_type == "vectorstores":
        return instantiate_vectorstore(class_object, params)
    elif base_type == "documentloaders":
        return instantiate_documentloader(node_type, class_object, params)
    elif base_type == "textsplitters":
        return instantiate_textsplitter(class_object, params)
    elif base_type == "utilities":
        return instantiate_utility(node_type, class_object, params)
    elif base_type == "chains":
        return instantiate_chains(node_type, class_object, params)
    elif base_type == "output_parsers":
        return instantiate_output_parser(node_type, class_object, params)
    elif base_type == "llms":
        return instantiate_llm(node_type, class_object, params)
    elif base_type == "retrievers":
        return instantiate_retriever(node_type, class_object, params)
    elif base_type == "memory":
        return instantiate_memory(node_type, class_object, params)
    elif base_type == "custom_components":
        return await instantiate_custom_component(node_type, class_object, params, user_id)
    elif base_type == "wrappers":
        return instantiate_wrapper(node_type, class_object, params)
    else:
        return class_object(**params)


async def instantiate_custom_component(node_type, class_object, params, user_id):
    params_copy = params.copy()
    class_object: "CustomComponent" = eval_custom_component_code(params_copy.pop("code"))
    custom_component = class_object(user_id=user_id)

    if "retriever" in params_copy and hasattr(params_copy["retriever"], "as_retriever"):
        params_copy["retriever"] = params_copy["retriever"].as_retriever()

    # Determine if the build method is asynchronous
    is_async = inspect.iscoroutinefunction(custom_component.build)

    if is_async:
        # Await the build method directly if it's async
        built_object = await custom_component.build(**params_copy)
    else:
        # Call the build method directly if it's sync
        built_object = custom_component.build(**params_copy)

    return built_object, {"repr": custom_component.custom_repr()}


def instantiate_wrapper(node_type, class_object, params):
    if node_type in wrapper_creator.from_method_nodes:
        method = wrapper_creator.from_method_nodes[node_type]
        if class_method := getattr(class_object, method, None):
            return class_method(**params)
        raise ValueError(f"Method {method} not found in {class_object}")
    return class_object(**params)


def instantiate_output_parser(node_type, class_object, params):
    if node_type in output_parser_creator.from_method_nodes:
        method = output_parser_creator.from_method_nodes[node_type]
        if class_method := getattr(class_object, method, None):
            return class_method(**params)
        raise ValueError(f"Method {method} not found in {class_object}")
    return class_object(**params)


def instantiate_llm(node_type, class_object, params: Dict):
    # This is a workaround so JinaChat works until streaming is implemented
    # if "openai_api_base" in params and "jina" in params["openai_api_base"]:
    # False if condition is True
    if "VertexAI" in node_type:
        return initialize_vertexai(class_object=class_object, params=params)
    # max_tokens sometimes is a string and should be an int
    if "max_tokens" in params:
        if isinstance(params["max_tokens"], str) and params["max_tokens"].isdigit():
            params["max_tokens"] = int(params["max_tokens"])
        elif not isinstance(params.get("max_tokens"), int):
            params.pop("max_tokens", None)
    return class_object(**params)


def instantiate_memory(node_type, class_object, params):
    # process input_key and output_key to remove them if
    # they are empty strings
    if node_type == "ConversationEntityMemory":
        params.pop("memory_key", None)

    for key in ["input_key", "output_key"]:
        if key in params and (params[key] == "" or not params[key]):
            params.pop(key)

    try:
        if "retriever" in params and hasattr(params["retriever"], "as_retriever"):
            params["retriever"] = params["retriever"].as_retriever()
        return class_object(**params)
    # I want to catch a specific attribute error that happens
    # when the object does not have a cursor attribute
    except Exception as exc:
        if "object has no attribute 'cursor'" in str(exc) or 'object has no field "conn"' in str(exc):
            raise AttributeError(
                (
                    "Failed to build connection to database."
                    f" Please check your connection string and try again. Error: {exc}"
                )
            ) from exc
        raise exc


def instantiate_retriever(node_type, class_object, params):
    if "retriever" in params and hasattr(params["retriever"], "as_retriever"):
        params["retriever"] = params["retriever"].as_retriever()
    if node_type in retriever_creator.from_method_nodes:
        method = retriever_creator.from_method_nodes[node_type]
        if class_method := getattr(class_object, method, None):
            return class_method(**params)
        raise ValueError(f"Method {method} not found in {class_object}")
    return class_object(**params)


def instantiate_chains(node_type, class_object: Type[Chain], params: Dict):
    from langflow.interface.chains.base import chain_creator

    if "retriever" in params and hasattr(params["retriever"], "as_retriever"):
        params["retriever"] = params["retriever"].as_retriever()
    if node_type in chain_creator.from_method_nodes:
        method = chain_creator.from_method_nodes[node_type]
        if class_method := getattr(class_object, method, None):
            return class_method(**params)
        raise ValueError(f"Method {method} not found in {class_object}")

    return class_object(**params)


def instantiate_agent(node_type, class_object: Type[agent_module.Agent], params: Dict):
    from langflow.interface.agents.base import agent_creator

    if node_type in agent_creator.from_method_nodes:
        method = agent_creator.from_method_nodes[node_type]
        if class_method := getattr(class_object, method, None):
            agent = class_method(**params)
            tools = params.get("tools", [])
            return AgentExecutor.from_agent_and_tools(agent=agent, tools=tools, handle_parsing_errors=True)
    return load_agent_executor(class_object, params)


def instantiate_prompt(node_type, class_object, params: Dict):
    params, prompt = handle_node_type(node_type, class_object, params)
    format_kwargs = handle_format_kwargs(prompt, params)
    # Now we'll use partial_format to format the prompt
    if format_kwargs:
        prompt = handle_partial_variables(prompt, format_kwargs)
    return prompt, format_kwargs


def instantiate_tool(node_type, class_object: Type[BaseTool], params: Dict):
    if node_type == "JsonSpec":
        if file_dict := load_file_into_dict(params.pop("path")):
            params["dict_"] = file_dict
        else:
            raise ValueError("Invalid file")
        return class_object(**params)
    elif node_type == "PythonFunctionTool":
        params["func"] = get_function(params.get("code"))
        return class_object(**params)
    elif node_type == "PythonFunction":
        function_string = params["code"]
        if isinstance(function_string, str):
            return validate.eval_function(function_string)
        raise ValueError("Function should be a string")
    elif node_type.lower() == "tool":
        return class_object(**params)
    return class_object(**params)


def instantiate_toolkit(node_type, class_object: Type[BaseToolkit], params: Dict):
    loaded_toolkit = class_object(**params)
    # Commenting this out for now to use toolkits as normal tools
    # if toolkits_creator.has_create_function(node_type):
    #     return load_toolkits_executor(node_type, loaded_toolkit, params)
    if isinstance(loaded_toolkit, BaseToolkit):
        return loaded_toolkit.get_tools()
    return loaded_toolkit


def instantiate_embedding(node_type, class_object, params: Dict):
    params.pop("model", None)
    params.pop("headers", None)

    if "VertexAI" in node_type:
        return initialize_vertexai(class_object=class_object, params=params)

    if "OpenAIEmbedding" in node_type:
        params["disallowed_special"] = ()

    try:
        return class_object(**params)
    except ValidationError:
        params = {key: value for key, value in params.items() if key in class_object.model_fields}
        return class_object(**params)


def instantiate_vectorstore(class_object: Type[VectorStore], params: Dict):
    search_kwargs = params.pop("search_kwargs", {})
    if search_kwargs == {"yourkey": "value"}:
        search_kwargs = {}
    # clean up docs or texts to have only documents
    if "texts" in params:
        params["documents"] = params.pop("texts")
    if "documents" in params:
        params["documents"] = [doc for doc in params["documents"] if isinstance(doc, Document)]
    if initializer := vecstore_initializer.get(class_object.__name__):
        vecstore = initializer(class_object, params)
    else:
        if "texts" in params:
            params["documents"] = params.pop("texts")
        vecstore = class_object.from_documents(**params)

    # ! This might not work. Need to test
    if search_kwargs and hasattr(vecstore, "as_retriever"):
        vecstore = vecstore.as_retriever(search_kwargs=search_kwargs)

    return vecstore


def instantiate_documentloader(node_type: str, class_object: Type[BaseLoader], params: Dict):
    if "file_filter" in params:
        # file_filter will be a string but we need a function
        # that will be used to filter the files using file_filter
        # like lambda x: x.endswith(".txt") but as we don't know
        # anything besides the string, we will simply check if the string is
        # in x and if it is, we will return True
        file_filter = params.pop("file_filter")
        extensions = file_filter.split(",")
        params["file_filter"] = lambda x: any(extension.strip() in x for extension in extensions)
    metadata = params.pop("metadata", None)
    if metadata and isinstance(metadata, str):
        try:
            metadata = orjson.loads(metadata)
        except json.JSONDecodeError as exc:
            raise ValueError("The metadata you provided is not a valid JSON string.") from exc

    if node_type == "WebBaseLoader":
        if web_path := params.pop("web_path", None):
            params["web_paths"] = [web_path]

    docs = class_object(**params).load()
    # Now if metadata is an empty dict, we will not add it to the documents
    if metadata:
        for doc in docs:
            # If the document already has metadata, we will not overwrite it
            if not doc.metadata:
                doc.metadata = metadata
            else:
                doc.metadata.update(metadata)

    return docs


def instantiate_textsplitter(
    class_object,
    params: Dict,
):
    try:
        documents = params.pop("documents")
        if not isinstance(documents, list):
            documents = [documents]
    except KeyError as exc:
        raise ValueError(
            "The source you provided did not load correctly or was empty."
            "Try changing the chunk_size of the Text Splitter."
        ) from exc

    if ("separator_type" in params and params["separator_type"] == "Text") or "separator_type" not in params:
        params.pop("separator_type", None)
        # separators might come in as an escaped string like \\n
        # so we need to convert it to a string
        if "separators" in params:
            params["separators"] = params["separators"].encode().decode("unicode-escape")
        text_splitter = class_object(**params)
    else:
        from langchain.text_splitter import Language

        language = params.pop("separator_type", None)
        params["language"] = Language(language)
        params.pop("separators", None)

        text_splitter = class_object.from_language(**params)
    return text_splitter.split_documents(documents)


def instantiate_utility(node_type, class_object, params: Dict):
    if node_type == "SQLDatabase":
        return class_object.from_uri(params.pop("uri"))
    return class_object(**params)


def replace_zero_shot_prompt_with_prompt_template(nodes):
    """Replace ZeroShotPrompt with PromptTemplate"""
    for node in nodes:
        if node["data"]["type"] == "ZeroShotPrompt":
            # Build Prompt Template
            tools = [
                tool
                for tool in nodes
                if tool["type"] != "chatOutputNode" and "Tool" in tool["data"]["node"]["base_classes"]
            ]
            node["data"] = build_prompt_template(prompt=node["data"], tools=tools)
            break
    return nodes


def load_agent_executor(agent_class: type[agent_module.Agent], params, **kwargs):
    """Load agent executor from agent class, tools and chain"""
    allowed_tools: Sequence[BaseTool] = params.get("allowed_tools", [])
    llm_chain = params["llm_chain"]
    # agent has hidden args for memory. might need to be support
    # memory = params["memory"]
    # if allowed_tools is not a list or set, make it a list
    if not isinstance(allowed_tools, (list, set)) and isinstance(allowed_tools, BaseTool):
        allowed_tools = [allowed_tools]
    tool_names = [tool.name for tool in allowed_tools]
    # Agent class requires an output_parser but Agent classes
    # have a default output_parser.
    agent = agent_class(allowed_tools=tool_names, llm_chain=llm_chain)  # type: ignore
    return AgentExecutor.from_agent_and_tools(
        agent=agent,
        tools=allowed_tools,
        handle_parsing_errors=True,
        # memory=memory,
        **kwargs,
    )


def load_toolkits_executor(node_type: str, toolkit: BaseToolkit, params: dict):
    create_function: Callable = toolkits_creator.get_create_function(node_type)
    if llm := params.get("llm"):
        return create_function(llm=llm, toolkit=toolkit)


def build_prompt_template(prompt, tools):
    """Build PromptTemplate from ZeroShotPrompt"""
    prefix = prompt["node"]["template"]["prefix"]["value"]
    suffix = prompt["node"]["template"]["suffix"]["value"]
    format_instructions = prompt["node"]["template"]["format_instructions"]["value"]

    tool_strings = "\n".join(
        [f"{tool['data']['node']['name']}: {tool['data']['node']['description']}" for tool in tools]
    )
    tool_names = ", ".join([tool["data"]["node"]["name"] for tool in tools])
    format_instructions = format_instructions.format(tool_names=tool_names)
    value = "\n\n".join([prefix, tool_strings, format_instructions, suffix])

    prompt["type"] = "PromptTemplate"

    prompt["node"] = {
        "template": {
            "_type": "prompt",
            "input_variables": {
                "type": "str",
                "required": True,
                "placeholder": "",
                "list": True,
                "show": False,
                "multiline": False,
            },
            "output_parser": {
                "type": "BaseOutputParser",
                "required": False,
                "placeholder": "",
                "list": False,
                "show": False,
                "multline": False,
                "value": None,
            },
            "template": {
                "type": "str",
                "required": True,
                "placeholder": "",
                "list": False,
                "show": True,
                "multiline": True,
                "value": value,
            },
            "template_format": {
                "type": "str",
                "required": False,
                "placeholder": "",
                "list": False,
                "show": False,
                "multline": False,
                "value": "f-string",
            },
            "validate_template": {
                "type": "bool",
                "required": False,
                "placeholder": "",
                "list": False,
                "show": False,
                "multline": False,
                "value": True,
            },
        },
        "description": "Schema to represent a prompt for an LLM.",
        "base_classes": ["BasePromptTemplate"],
    }

    return prompt



from typing import Optional

from langchain_google_genai import ChatGoogleGenerativeAI  # type: ignore
from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel, RangeSpec, TemplateField
from pydantic.v1.types import SecretStr


class GoogleGenerativeAIComponent(CustomComponent):
    display_name: str = "Google Generative AI"
    description: str = "A component that uses Google Generative AI to generate text."
    documentation: str = "http://docs.langflow.org/components/custom"

    def build_config(self):
        return {
            "google_api_key": TemplateField(
                display_name="Google API Key",
                info="The Google API Key to use for the Google Generative AI.",
            ),
            "max_output_tokens": TemplateField(
                display_name="Max Output Tokens",
                info="The maximum number of tokens to generate.",
            ),
            "temperature": TemplateField(
                display_name="Temperature",
                info="Run inference with this temperature. Must by in the closed interval [0.0, 1.0].",
            ),
            "top_k": TemplateField(
                display_name="Top K",
                info="Decode using top-k sampling: consider the set of top_k most probable tokens. Must be positive.",
                range_spec=RangeSpec(min=0, max=2, step=0.1),
                advanced=True,
            ),
            "top_p": TemplateField(
                display_name="Top P",
                info="The maximum cumulative probability of tokens to consider when sampling.",
                advanced=True,
            ),
            "n": TemplateField(
                display_name="N",
                info="Number of chat completions to generate for each prompt. Note that the API may not return the full n completions if duplicates are generated.",
                advanced=True,
            ),
            "model": TemplateField(
                display_name="Model",
                info="The name of the model to use. Supported examples: gemini-pro",
                options=["gemini-pro", "gemini-pro-vision"],
            ),
            "code": TemplateField(
                advanced=True,
            ),
        }

    def build(
        self,
        google_api_key: str,
        model: str,
        max_output_tokens: Optional[int] = None,
        temperature: float = 0.1,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
        n: Optional[int] = 1,
    ) -> BaseLanguageModel:
        return ChatGoogleGenerativeAI(
            model=model,
            max_output_tokens=max_output_tokens or None,  # type: ignore
            temperature=temperature,
            top_k=top_k or None,
            top_p=top_p or None,  # type: ignore
            n=n or 1,
            google_api_key=SecretStr(google_api_key),
        )



from typing import TYPE_CHECKING, Optional

from langflow.interface.run import build_sorted_vertices
from langflow.services.base import Service
from langflow.services.session.utils import compute_dict_hash, session_id_generator

if TYPE_CHECKING:
    from langflow.services.cache.base import BaseCacheService


class SessionService(Service):
    name = "session_service"

    def __init__(self, cache_service):
        self.cache_service: "BaseCacheService" = cache_service

    async def load_session(self, key, data_graph: Optional[dict] = None):
        # Check if the data is cached
        if key in self.cache_service:
            return self.cache_service.get(key)

        if key is None:
            key = self.generate_key(session_id=None, data_graph=data_graph)
        if data_graph is None:
            return (None, None)
        # If not cached, build the graph and cache it
        graph, artifacts = await build_sorted_vertices(data_graph)

        self.cache_service.set(key, (graph, artifacts))

        return graph, artifacts

    def build_key(self, session_id, data_graph):
        json_hash = compute_dict_hash(data_graph)
        return f"{session_id}{':' if session_id else ''}{json_hash}"

    def generate_key(self, session_id, data_graph):
        # Hash the JSON and combine it with the session_id to create a unique key
        if session_id is None:
            # generate a 5 char session_id to concatenate with the json_hash
            session_id = session_id_generator()
        return self.build_key(session_id, data_graph=data_graph)

    def update_session(self, session_id, value):
        self.cache_service.set(session_id, value)

    def clear_session(self, session_id):
        self.cache_service.delete(session_id)



import pickle
import threading
import time
from collections import OrderedDict

from loguru import logger

from langflow.services.base import Service
from langflow.services.cache.base import BaseCacheService


class InMemoryCache(BaseCacheService, Service):

    """
    A simple in-memory cache using an OrderedDict.

    This cache supports setting a maximum size and expiration time for cached items.
    When the cache is full, it uses a Least Recently Used (LRU) eviction policy.
    Thread-safe using a threading Lock.

    Attributes:
        max_size (int, optional): Maximum number of items to store in the cache.
        expiration_time (int, optional): Time in seconds after which a cached item expires. Default is 1 hour.

    Example:

        cache = InMemoryCache(max_size=3, expiration_time=5)

        # setting cache values
        cache.set("a", 1)
        cache.set("b", 2)
        cache["c"] = 3

        # getting cache values
        a = cache.get("a")
        b = cache["b"]
    """

    def __init__(self, max_size=None, expiration_time=60 * 60):
        """
        Initialize a new InMemoryCache instance.

        Args:
            max_size (int, optional): Maximum number of items to store in the cache.
            expiration_time (int, optional): Time in seconds after which a cached item expires. Default is 1 hour.
        """
        self._cache = OrderedDict()
        self._lock = threading.RLock()
        self.max_size = max_size
        self.expiration_time = expiration_time

    def get(self, key):
        """
        Retrieve an item from the cache.

        Args:
            key: The key of the item to retrieve.

        Returns:
            The value associated with the key, or None if the key is not found or the item has expired.
        """
        with self._lock:
            return self._get_without_lock(key)

    def _get_without_lock(self, key):
        """
        Retrieve an item from the cache without acquiring the lock.
        """
        if item := self._cache.get(key):
            if self.expiration_time is None or time.time() - item["time"] < self.expiration_time:
                # Move the key to the end to make it recently used
                self._cache.move_to_end(key)
                # Check if the value is pickled
                if isinstance(item["value"], bytes):
                    value = pickle.loads(item["value"])
                else:
                    value = item["value"]
                return value
            else:
                self.delete(key)
        return None

    def set(self, key, value, pickle=False):
        """
        Add an item to the cache.

        If the cache is full, the least recently used item is evicted.

        Args:
            key: The key of the item.
            value: The value to cache.
        """
        with self._lock:
            if key in self._cache:
                # Remove existing key before re-inserting to update order
                self.delete(key)
            elif self.max_size and len(self._cache) >= self.max_size:
                # Remove least recently used item
                self._cache.popitem(last=False)
            # pickle locally to mimic Redis
            if pickle:
                value = pickle.dumps(value)

            self._cache[key] = {"value": value, "time": time.time()}

    def upsert(self, key, value):
        """
        Inserts or updates a value in the cache.
        If the existing value and the new value are both dictionaries, they are merged.

        Args:
            key: The key of the item.
            value: The value to insert or update.
        """
        with self._lock:
            existing_value = self._get_without_lock(key)
            if existing_value is not None and isinstance(existing_value, dict) and isinstance(value, dict):
                existing_value.update(value)
                value = existing_value

            self.set(key, value)

    def get_or_set(self, key, value):
        """
        Retrieve an item from the cache. If the item does not exist,
        set it with the provided value.

        Args:
            key: The key of the item.
            value: The value to cache if the item doesn't exist.

        Returns:
            The cached value associated with the key.
        """
        with self._lock:
            if key in self._cache:
                return self.get(key)
            self.set(key, value)
            return value

    def delete(self, key):
        """
        Remove an item from the cache.

        Args:
            key: The key of the item to remove.
        """
        with self._lock:
            self._cache.pop(key, None)

    def clear(self):
        """
        Clear all items from the cache.
        """
        with self._lock:
            self._cache.clear()

    def __contains__(self, key):
        """Check if the key is in the cache."""
        return key in self._cache

    def __getitem__(self, key):
        """Retrieve an item from the cache using the square bracket notation."""
        return self.get(key)

    def __setitem__(self, key, value):
        """Add an item to the cache using the square bracket notation."""
        self.set(key, value)

    def __delitem__(self, key):
        """Remove an item from the cache using the square bracket notation."""
        self.delete(key)

    def __len__(self):
        """Return the number of items in the cache."""
        return len(self._cache)

    def __repr__(self):
        """Return a string representation of the InMemoryCache instance."""
        return f"InMemoryCache(max_size={self.max_size}, expiration_time={self.expiration_time})"


class RedisCache(BaseCacheService, Service):
    """
    A Redis-based cache implementation.

    This cache supports setting an expiration time for cached items.

    Attributes:
        expiration_time (int, optional): Time in seconds after which a cached item expires. Default is 1 hour.

    Example:

        cache = RedisCache(expiration_time=5)

        # setting cache values
        cache.set("a", 1)
        cache.set("b", 2)
        cache["c"] = 3

        # getting cache values
        a = cache.get("a")
        b = cache["b"]
    """

    def __init__(self, host="localhost", port=6379, db=0, url=None, expiration_time=60 * 60):
        """
        Initialize a new RedisCache instance.

        Args:
            host (str, optional): Redis host.
            port (int, optional): Redis port.
            db (int, optional): Redis DB.
            expiration_time (int, optional): Time in seconds after which a
            ached item expires. Default is 1 hour.
        """
        try:
            import redis
        except ImportError as exc:
            raise ImportError(
                "RedisCache requires the redis-py package."
                " Please install Langflow with the deploy extra: pip install langflow[deploy]"
            ) from exc
        logger.warning(
            "RedisCache is an experimental feature and may not work as expected."
            " Please report any issues to our GitHub repository."
        )
        if url:
            self._client = redis.StrictRedis.from_url(url)
        else:
            self._client = redis.StrictRedis(host=host, port=port, db=db)
        self.expiration_time = expiration_time

    # check connection
    def is_connected(self):
        """
        Check if the Redis client is connected.
        """
        import redis

        try:
            self._client.ping()
            return True
        except redis.exceptions.ConnectionError:
            return False

    def get(self, key):
        """
        Retrieve an item from the cache.

        Args:
            key: The key of the item to retrieve.

        Returns:
            The value associated with the key, or None if the key is not found.
        """
        value = self._client.get(key)
        return pickle.loads(value) if value else None

    def set(self, key, value):
        """
        Add an item to the cache.

        Args:
            key: The key of the item.
            value: The value to cache.
        """
        try:
            if pickled := pickle.dumps(value):
                result = self._client.setex(key, self.expiration_time, pickled)
                if not result:
                    raise ValueError("RedisCache could not set the value.")
        except TypeError as exc:
            raise TypeError("RedisCache only accepts values that can be pickled. ") from exc

    def upsert(self, key, value):
        """
        Inserts or updates a value in the cache.
        If the existing value and the new value are both dictionaries, they are merged.

        Args:
            key: The key of the item.
            value: The value to insert or update.
        """
        existing_value = self.get(key)
        if existing_value is not None and isinstance(existing_value, dict) and isinstance(value, dict):
            existing_value.update(value)
            value = existing_value

        self.set(key, value)

    def delete(self, key):
        """
        Remove an item from the cache.

        Args:
            key: The key of the item to remove.
        """
        self._client.delete(key)

    def clear(self):
        """
        Clear all items from the cache.
        """
        self._client.flushdb()

    def __contains__(self, key):
        """Check if the key is in the cache."""
        return False if key is None else self._client.exists(key)

    def __getitem__(self, key):
        """Retrieve an item from the cache using the square bracket notation."""
        return self.get(key)

    def __setitem__(self, key, value):
        """Add an item to the cache using the square bracket notation."""
        self.set(key, value)

    def __delitem__(self, key):
        """Remove an item from the cache using the square bracket notation."""
        self.delete(key)

    def __repr__(self):
        """Return a string representation of the RedisCache instance."""
        return f"RedisCache(expiration_time={self.expiration_time})"



from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langflow.template.template.base import Template
from langflow.utils.constants import DEFAULT_PYTHON_FUNCTION


class ToolNode(FrontendNode):
    name: str = "Tool"
    template: Template = Template(
        type_name="Tool",
        fields=[
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=True,
                value="",
                name="name",
                advanced=False,
            ),
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=True,
                value="",
                name="description",
                advanced=False,
            ),
            TemplateField(
                name="func",
                field_type="Callable",
                required=True,
                is_list=False,
                show=True,
                multiline=True,
                advanced=False,
            ),
            TemplateField(
                field_type="bool",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=False,
                value=False,
                name="return_direct",
            ),
        ],
    )
    description: str = "Converts a chain, agent or function into a tool."
    base_classes: list[str] = ["Tool", "BaseTool"]


class PythonFunctionToolNode(FrontendNode):
    name: str = "PythonFunctionTool"
    template: Template = Template(
        type_name="PythonFunctionTool",
        fields=[
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=False,
                value="",
                name="name",
                advanced=False,
            ),
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=False,
                value="",
                name="description",
                advanced=False,
            ),
            TemplateField(
                field_type="code",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                value=DEFAULT_PYTHON_FUNCTION,
                name="code",
                advanced=False,
            ),
            TemplateField(
                field_type="bool",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=False,
                value=False,
                name="return_direct",
            ),
        ],
    )
    description: str = "Python function to be executed."
    base_classes: list[str] = ["BaseTool", "Tool"]


class PythonFunctionNode(FrontendNode):
    name: str = "PythonFunction"
    template: Template = Template(
        type_name="PythonFunction",
        fields=[
            TemplateField(
                field_type="code",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                value=DEFAULT_PYTHON_FUNCTION,
                name="code",
                advanced=False,
            )
        ],
    )
    description: str = "Python function to be executed."
    base_classes: list[str] = ["Callable"]



from typing import Optional

from langflow.services.database.models.base import orjson_dumps
from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langflow.template.frontend_node.constants import CTRANSFORMERS_DEFAULT_CONFIG, OPENAI_API_BASE_INFO


class LLMFrontendNode(FrontendNode):
    def add_extra_fields(self) -> None:
        if "VertexAI" in self.template.type_name:
            # Add credentials field which should of type file.
            self.template.add_field(
                TemplateField(
                    field_type="file",
                    required=False,
                    show=True,
                    name="credentials",
                    value="",
                    file_types=[".json"],
                )
            )

    @staticmethod
    def format_vertex_field(field: TemplateField, name: str):
        key = field.name or ""
        if "VertexAI" in name:
            advanced_fields = [
                "tuned_model_name",
                "verbose",
                "top_p",
                "top_k",
                "max_output_tokens",
            ]
            if key in advanced_fields:
                field.advanced = True
            show_fields = [
                "tuned_model_name",
                "verbose",
                "project",
                "location",
                "credentials",
                "max_output_tokens",
                "model_name",
                "temperature",
                "top_p",
                "top_k",
            ]

            if key in show_fields:
                field.show = True

    @staticmethod
    def format_openai_field(field: TemplateField):
        key = field.name or ""
        if "openai" in key.lower():
            field.display_name = (key.title().replace("Openai", "OpenAI").replace("_", " ")).replace("Api", "API")

        if "key" not in key.lower() and "token" not in key.lower():
            field.password = False

        if key == "openai_api_base":
            field.info = OPENAI_API_BASE_INFO

    def add_extra_base_classes(self) -> None:
        if "BaseLLM" not in self.base_classes:
            self.base_classes.append("BaseLLM")

    @staticmethod
    def format_azure_field(field: TemplateField):
        key = field.name or ""
        if key == "model_name":
            field.show = False  # Azure uses deployment_name instead of model_name.
        elif key == "openai_api_type":
            field.show = False
            field.password = False
            field.value = "azure"
        elif key == "openai_api_version":
            field.password = False

    @staticmethod
    def format_llama_field(field: TemplateField):
        field.show = True
        field.advanced = not field.required

    @staticmethod
    def format_ctransformers_field(field: TemplateField):
        key = field.name or ""
        if key == "config":
            field.show = True
            field.advanced = True
            field.value = orjson_dumps(CTRANSFORMERS_DEFAULT_CONFIG, indent_2=True)

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        display_names_dict = {
            "huggingfacehub_api_token": "HuggingFace Hub API Token",
        }
        FrontendNode.format_field(field, name)
        LLMFrontendNode.format_openai_field(field)
        LLMFrontendNode.format_ctransformers_field(field)
        if name and "azure" in name.lower():
            LLMFrontendNode.format_azure_field(field)
        if name and "llama" in name.lower():
            LLMFrontendNode.format_llama_field(field)
        if name and "vertex" in name.lower():
            LLMFrontendNode.format_vertex_field(field, name)
        SHOW_FIELDS = ["repo_id"]
        key = field.name or ""
        if key in SHOW_FIELDS:
            field.show = True

        if "api" in key and ("key" in key or ("token" in key and "tokens" not in key)):
            field.password = True
            field.show = True
            # Required should be False to support
            # loading the API key from environment variables
            field.required = False
            field.advanced = False

        if key == "task":
            field.required = True
            field.show = True
            field.is_list = True
            field.options = ["text-generation", "text2text-generation", "summarization"]
            field.value = field.options[0]
            field.advanced = True

        if display_name := display_names_dict.get(key):
            field.display_name = display_name
        if key == "model_kwargs":
            field.field_type = "dict"
            field.advanced = True
            field.show = True
        elif key in [
            "model_name",
            "temperature",
            "model_file",
            "model_type",
            "deployment_name",
            "credentials",
        ]:
            field.advanced = False
            field.show = True
        if key == "credentials":
            field.field_type = "file"
        if name == "VertexAI" and key not in [
            "callbacks",
            "client",
            "stop",
            "tags",
            "cache",
        ]:
            field.show = True



from datetime import datetime
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException
from langflow.services.auth import utils as auth_utils
from langflow.services.auth.utils import get_current_active_user
from langflow.services.database.models.credential import Credential, CredentialCreate, CredentialRead, CredentialUpdate
from langflow.services.database.models.user.model import User
from langflow.services.deps import get_session, get_settings_service
from sqlmodel import Session, select

router = APIRouter(prefix="/credentials", tags=["Credentials"])


@router.post("/", response_model=CredentialRead, status_code=201)
def create_credential(
    *,
    session: Session = Depends(get_session),
    credential: CredentialCreate,
    current_user: User = Depends(get_current_active_user),
    settings_service=Depends(get_settings_service),
):
    """Create a new credential."""
    try:
        # check if credential name already exists
        credential_exists = session.exec(
            select(Credential).where(Credential.name == credential.name, Credential.user_id == current_user.id)
        ).first()
        if credential_exists:
            raise HTTPException(status_code=400, detail="Credential name already exists")

        db_credential = Credential.model_validate(credential, from_attributes=True)
        if not db_credential.value:
            raise HTTPException(status_code=400, detail="Credential value cannot be empty")
        encrypted = auth_utils.encrypt_api_key(db_credential.value, settings_service=settings_service)
        db_credential.value = encrypted
        db_credential.user_id = current_user.id
        session.add(db_credential)
        session.commit()
        session.refresh(db_credential)
        return db_credential
    except Exception as e:
        if isinstance(e, HTTPException):
            raise e
        raise HTTPException(status_code=500, detail=str(e)) from e


@router.get("/", response_model=list[CredentialRead], status_code=200)
def read_credentials(
    *,
    session: Session = Depends(get_session),
    current_user: User = Depends(get_current_active_user),
):
    """Read all credentials."""
    try:
        credentials = session.exec(select(Credential).where(Credential.user_id == current_user.id)).all()
        return credentials
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e)) from e


@router.patch("/{credential_id}", response_model=CredentialRead, status_code=200)
def update_credential(
    *,
    session: Session = Depends(get_session),
    credential_id: UUID,
    credential: CredentialUpdate,
    current_user: User = Depends(get_current_active_user),
):
    """Update a credential."""
    try:
        db_credential = session.exec(
            select(Credential).where(Credential.id == credential_id, Credential.user_id == current_user.id)
        ).first()
        if not db_credential:
            raise HTTPException(status_code=404, detail="Credential not found")

        credential_data = credential.model_dump(exclude_unset=True)
        for key, value in credential_data.items():
            setattr(db_credential, key, value)
        db_credential.updated_at = datetime.utcnow()
        session.commit()
        session.refresh(db_credential)
        return db_credential
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e)) from e



import time

from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    WebSocket,
    WebSocketException,
    status,
)
from fastapi.responses import StreamingResponse
from loguru import logger
from sqlmodel import Session

from langflow.api.utils import build_input_keys_response, format_elapsed_time
from langflow.api.v1.schemas import BuildStatus, BuiltResponse, InitResponse, StreamData
from langflow.graph.graph.base import Graph
from langflow.services.auth.utils import (
    get_current_active_user,
    get_current_user_for_websocket,
)
from langflow.services.cache.service import BaseCacheService
from langflow.services.cache.utils import update_build_status
from langflow.services.chat.service import ChatService
from langflow.services.deps import get_cache_service, get_chat_service, get_session

router = APIRouter(tags=["Chat"])


@router.websocket("/chat/{client_id}")
async def chat(
    client_id: str,
    websocket: WebSocket,
    db: Session = Depends(get_session),
    chat_service: "ChatService" = Depends(get_chat_service),
):
    """Websocket endpoint for chat."""
    try:
        user = await get_current_user_for_websocket(websocket, db)
        await websocket.accept()
        if not user:
            await websocket.close(code=status.WS_1008_POLICY_VIOLATION, reason="Unauthorized")
        elif not user.is_active:
            await websocket.close(code=status.WS_1008_POLICY_VIOLATION, reason="Unauthorized")

        if client_id in chat_service.cache_service:
            await chat_service.handle_websocket(client_id, websocket)
        else:
            # We accept the connection but close it immediately
            # if the flow is not built yet
            message = "Please, build the flow before sending messages"
            await websocket.close(code=status.WS_1011_INTERNAL_ERROR, reason=message)
    except WebSocketException as exc:
        logger.error(f"Websocket exrror: {exc}")
        await websocket.close(code=status.WS_1011_INTERNAL_ERROR, reason=str(exc))
    except Exception as exc:
        logger.error(f"Error in chat websocket: {exc}")
        messsage = exc.detail if isinstance(exc, HTTPException) else str(exc)
        if "Could not validate credentials" in str(exc):
            await websocket.close(code=status.WS_1008_POLICY_VIOLATION, reason="Unauthorized")
        else:
            await websocket.close(code=status.WS_1011_INTERNAL_ERROR, reason=messsage)


@router.post("/build/init/{flow_id}", response_model=InitResponse, status_code=201)
async def init_build(
    graph_data: dict,
    flow_id: str,
    current_user=Depends(get_current_active_user),
    chat_service: "ChatService" = Depends(get_chat_service),
    cache_service: "BaseCacheService" = Depends(get_cache_service),
):
    """Initialize the build by storing graph data and returning a unique session ID."""
    try:
        if flow_id is None:
            raise ValueError("No ID provided")
        # Check if already building
        if (
            flow_id in cache_service
            and isinstance(cache_service[flow_id], dict)
            and cache_service[flow_id].get("status") == BuildStatus.IN_PROGRESS
        ):
            return InitResponse(flowId=flow_id)

        # Delete from cache if already exists
        if flow_id in chat_service.cache_service:
            chat_service.cache_service.delete(flow_id)
            logger.debug(f"Deleted flow {flow_id} from cache")
        cache_service[flow_id] = {
            "graph_data": graph_data,
            "status": BuildStatus.STARTED,
            "user_id": current_user.id,
        }

        return InitResponse(flowId=flow_id)
    except Exception as exc:
        logger.error(f"Error initializing build: {exc}")
        return HTTPException(status_code=500, detail=str(exc))


@router.get("/build/{flow_id}/status", response_model=BuiltResponse)
async def build_status(flow_id: str, cache_service: "BaseCacheService" = Depends(get_cache_service)):
    """Check the flow_id is in the cache_service."""
    try:
        built = flow_id in cache_service and cache_service[flow_id]["status"] == BuildStatus.SUCCESS

        return BuiltResponse(
            built=built,
        )

    except Exception as exc:
        logger.error(f"Error checking build status: {exc}")
        return HTTPException(status_code=500, detail=str(exc))


@router.get("/build/stream/{flow_id}", response_class=StreamingResponse)
async def stream_build(
    flow_id: str,
    chat_service: "ChatService" = Depends(get_chat_service),
    cache_service: "BaseCacheService" = Depends(get_cache_service),
):
    """Stream the build process based on stored flow data."""

    async def event_stream(flow_id):
        final_response = {"end_of_stream": True}
        artifacts = {}
        flow_cache = cache_service[flow_id]
        flow_cache = flow_cache if isinstance(flow_cache, dict) else {}
        try:
            if flow_id not in cache_service:
                error_message = "Invalid session ID"
                yield str(StreamData(event="error", data={"error": error_message}))
                return

            if flow_cache.get("status") == BuildStatus.IN_PROGRESS:
                error_message = "Already building"
                yield str(StreamData(event="error", data={"error": error_message}))
                return

            graph_data = flow_cache.get("graph_data")

            if not graph_data:
                error_message = "No data provided"
                yield str(StreamData(event="error", data={"error": error_message}))
                return

            logger.debug("Building langchain object")

            # Some error could happen when building the graph
            graph = Graph.from_payload(graph_data)

            number_of_nodes = len(graph.vertices)
            update_build_status(cache_service, flow_id, BuildStatus.IN_PROGRESS)
            time_elapsed = ""
            try:
                user_id = flow_cache["user_id"]
            except KeyError:
                logger.debug("No user_id found in cache_service")
                user_id = None
            for i, vertex in enumerate(graph.generator_build(), 1):
                start_time = time.perf_counter()
                try:
                    log_dict = {
                        "log": f"Building node {vertex.vertex_type}",
                    }
                    yield str(StreamData(event="log", data=log_dict))
                    if vertex.is_task:
                        vertex = await try_running_celery_task(vertex, user_id)
                    else:
                        await vertex.build(user_id=user_id)
                    time_elapsed = format_elapsed_time(time.perf_counter() - start_time)
                    params = vertex._built_object_repr()
                    valid = True

                    logger.debug(f"Building node {str(vertex.vertex_type)}")
                    logger.debug(f"Output: {params[:100]}{'...' if len(params) > 100 else ''}")
                    if vertex.artifacts:
                        # The artifacts will be prompt variables
                        # passed to build_input_keys_response
                        # to set the input_keys values
                        artifacts.update(vertex.artifacts)
                except Exception as exc:
                    logger.exception(exc)
                    params = str(exc)
                    valid = False
                    time_elapsed = format_elapsed_time(time.perf_counter() - start_time)
                    update_build_status(cache_service, flow_id, BuildStatus.FAILURE)

                vertex_id = vertex.parent_node_id if vertex.parent_is_top_level else vertex.id
                if vertex_id in graph.top_level_vertices:
                    response = {
                        "valid": valid,
                        "params": params,
                        "id": vertex_id,
                        "progress": round(i / number_of_nodes, 2),
                        "duration": time_elapsed,
                    }

                    yield str(StreamData(event="message", data=response))

            langchain_object = await graph.build()
            # Now we  need to check the input_keys to send them to the client
            if hasattr(langchain_object, "input_keys"):
                input_keys_response = build_input_keys_response(langchain_object, artifacts)
            else:
                input_keys_response = {
                    "input_keys": None,
                    "memory_keys": [],
                    "handle_keys": [],
                }
            yield str(StreamData(event="message", data=input_keys_response))
            chat_service.set_cache(flow_id, langchain_object)
            # We need to reset the chat history
            chat_service.chat_history.empty_history(flow_id)
            update_build_status(cache_service, flow_id, BuildStatus.SUCCESS)
        except Exception as exc:
            logger.exception(exc)
            logger.error("Error while building the flow: %s", exc)

            update_build_status(cache_service, flow_id, BuildStatus.FAILURE)
            yield str(StreamData(event="error", data={"error": str(exc)}))
        finally:
            yield str(StreamData(event="message", data=final_response))

    try:
        return StreamingResponse(event_stream(flow_id), media_type="text/event-stream")
    except Exception as exc:
        logger.error(f"Error streaming build: {exc}")
        raise HTTPException(status_code=500, detail=str(exc))


async def try_running_celery_task(vertex, user_id):
    # Try running the task in celery
    # and set the task_id to the local vertex
    # if it fails, run the task locally
    try:
        from langflow.worker import build_vertex

        task = build_vertex.delay(vertex)
        vertex.task_id = task.id
    except Exception as exc:
        logger.debug(f"Error running task in celery: {exc}")
        vertex.task_id = None
        await vertex.build(user_id=user_id)
    return vertex



from typing import Annotated, List, Optional, Union
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query

from langflow.api.utils import check_langflow_version
from langflow.services.auth import utils as auth_utils
from langflow.services.database.models.user.model import User
from langflow.services.deps import get_settings_service, get_store_service
from langflow.services.store.exceptions import CustomException
from langflow.services.store.schema import (
    CreateComponentResponse,
    DownloadComponentResponse,
    ListComponentResponseModel,
    StoreComponentCreate,
    TagResponse,
    UsersLikesResponse,
)
from langflow.services.store.service import StoreService

router = APIRouter(prefix="/store", tags=["Components Store"])


def get_user_store_api_key(
    user: User = Depends(auth_utils.get_current_active_user),
    settings_service=Depends(get_settings_service),
):
    if not user.store_api_key:
        raise HTTPException(status_code=400, detail="You must have a store API key set.")
    decrypted = auth_utils.decrypt_api_key(user.store_api_key, settings_service)
    return decrypted


def get_optional_user_store_api_key(
    user: User = Depends(auth_utils.get_current_active_user),
    settings_service=Depends(get_settings_service),
):
    if not user.store_api_key:
        return None
    decrypted = auth_utils.decrypt_api_key(user.store_api_key, settings_service)
    return decrypted


@router.get("/check/")
def check_if_store_is_enabled(
    settings_service=Depends(get_settings_service),
):
    return {
        "enabled": settings_service.settings.STORE,
    }


@router.get("/check/api_key")
async def check_if_store_has_api_key(
    api_key: Optional[str] = Depends(get_optional_user_store_api_key),
    store_service: StoreService = Depends(get_store_service),
):
    if api_key is None:
        return {"has_api_key": False, "is_valid": False}

    try:
        is_valid = await store_service.check_api_key(api_key)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))

    return {"has_api_key": api_key is not None, "is_valid": is_valid}


@router.post("/components/", response_model=CreateComponentResponse, status_code=201)
async def share_component(
    component: StoreComponentCreate,
    store_service: StoreService = Depends(get_store_service),
    store_api_key: str = Depends(get_user_store_api_key),
):
    try:
        await check_langflow_version(component)
        result = await store_service.upload(store_api_key, component)
        return result
    except Exception as exc:
        raise HTTPException(status_code=400, detail=str(exc))


@router.patch("/components/{component_id}", response_model=CreateComponentResponse, status_code=201)
async def update_shared_component(
    component_id: UUID,
    component: StoreComponentCreate,
    store_service: StoreService = Depends(get_store_service),
    store_api_key: str = Depends(get_user_store_api_key),
):
    try:
        await check_langflow_version(component)
        result = await store_service.update(store_api_key, component_id, component)
        return result
    except Exception as exc:
        raise HTTPException(status_code=400, detail=str(exc))


@router.get("/components/", response_model=ListComponentResponseModel)
async def get_components(
    component_id: Annotated[Optional[str], Query()] = None,
    search: Annotated[Optional[str], Query()] = None,
    private: Annotated[Optional[bool], Query()] = None,
    is_component: Annotated[Optional[bool], Query()] = None,
    tags: Annotated[Optional[list[str]], Query()] = None,
    sort: Annotated[Union[list[str], None], Query()] = None,
    liked: Annotated[bool, Query()] = False,
    filter_by_user: Annotated[bool, Query()] = False,
    fields: Annotated[Optional[list[str]], Query()] = None,
    page: int = 1,
    limit: int = 10,
    store_service: StoreService = Depends(get_store_service),
    store_api_key: Optional[str] = Depends(get_optional_user_store_api_key),
):
    try:
        return await store_service.get_list_component_response_model(
            component_id=component_id,
            search=search,
            private=private,
            is_component=is_component,
            fields=fields,
            tags=tags,
            sort=sort,
            liked=liked,
            filter_by_user=filter_by_user,
            page=page,
            limit=limit,
            store_api_key=store_api_key,
        )
    except CustomException as exc:
        raise HTTPException(status_code=exc.status_code, detail=str(exc)) from exc
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc)) from exc


@router.get("/components/{component_id}", response_model=DownloadComponentResponse)
async def download_component(
    component_id: UUID,
    store_service: StoreService = Depends(get_store_service),
    store_api_key: str = Depends(get_user_store_api_key),
):
    try:
        component = await store_service.download(store_api_key, component_id)
    except CustomException as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc)) from exc

    if component is None:
        raise HTTPException(status_code=400, detail="Component not found")

    return component


@router.get("/tags", response_model=List[TagResponse])
async def get_tags(
    store_service: StoreService = Depends(get_store_service),
):
    try:
        return await store_service.get_tags()
    except CustomException as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc))


@router.get("/users/likes", response_model=List[UsersLikesResponse])
async def get_list_of_components_liked_by_user(
    store_service: StoreService = Depends(get_store_service),
    store_api_key: str = Depends(get_user_store_api_key),
):
    try:
        return await store_service.get_user_likes(store_api_key)
    except CustomException as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc))


@router.post("/users/likes/{component_id}", response_model=UsersLikesResponse)
async def like_component(
    component_id: UUID,
    store_service: StoreService = Depends(get_store_service),
    store_api_key: str = Depends(get_user_store_api_key),
):
    try:
        result = await store_service.like_component(store_api_key, str(component_id))
        likes_count = await store_service.get_component_likes_count(str(component_id), store_api_key)

        return UsersLikesResponse(likes_count=likes_count, liked_by_user=result)
    except CustomException as exc:
        raise HTTPException(status_code=exc.status_code, detail=str(exc)) from exc
    except Exception as exc:
        raise HTTPException(status_code=500, detail=str(exc))



import os
from pathlib import Path
from typing import Optional

import orjson
from loguru import logger
from platformdirs import user_cache_dir
from rich.logging import RichHandler

VALID_LOG_LEVELS = ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]


def serialize(record):
    subset = {
        "timestamp": record["time"].timestamp(),
        "message": record["message"],
        "level": record["level"].name,
        "module": record["module"],
    }
    return orjson.dumps(subset)


def patching(record):
    record["extra"]["serialized"] = serialize(record)


def configure(log_level: Optional[str] = None, log_file: Optional[Path] = None):
    if os.getenv("LANGFLOW_LOG_LEVEL") in VALID_LOG_LEVELS and log_level is None:
        log_level = os.getenv("LANGFLOW_LOG_LEVEL")
    if log_level is None:
        log_level = "INFO"
    # Human-readable
    log_format = (
        "<green>{time:YYYY-MM-DD HH:mm:ss}</green> - <level>"
        "{level: <8}</level> - {module} - <level>{message}</level>"
    )

    # log_format = log_format_dev if log_level.upper() == "DEBUG" else log_format_prod
    logger.remove()  # Remove default handlers
    logger.patch(patching)
    # Configure loguru to use RichHandler
    logger.configure(
        handlers=[
            {
                "sink": RichHandler(rich_tracebacks=True, markup=True),
                "format": log_format,
                "level": log_level.upper(),
            }
        ]
    )

    if not log_file:
        cache_dir = Path(user_cache_dir("langflow"))
        log_file = cache_dir / "langflow.log"

    log_file = Path(log_file)
    log_file.parent.mkdir(parents=True, exist_ok=True)

    logger.add(
        sink=str(log_file),
        level=log_level.upper(),
        format=log_format,
        rotation="10 MB",  # Log rotation based on file size
        serialize=True,
    )

    logger.debug(f"Logger set up with log level: {log_level}")
    if log_file:
        logger.debug(f"Log file: {log_file}")



from typing import Dict, List, Optional

from langchain.agents.load_tools import (
    _EXTRA_LLM_TOOLS,
    _EXTRA_OPTIONAL_TOOLS,
    _LLM_TOOLS,
)

from langflow.custom import customs
from langflow.interface.base import LangChainTypeCreator
from langflow.interface.tools.constants import (
    ALL_TOOLS_NAMES,
    CUSTOM_TOOLS,
    FILE_TOOLS,
    OTHER_TOOLS,
)
from langflow.interface.tools.util import get_tool_params
from langflow.services.deps import get_settings_service

from langflow.template.field.base import TemplateField
from langflow.template.template.base import Template
from langflow.utils import util
from langflow.utils.util import build_template_from_class
from langflow.utils.logger import logger

TOOL_INPUTS = {
    "str": TemplateField(
        field_type="str",
        required=True,
        is_list=False,
        show=True,
        placeholder="",
        value="",
    ),
    "llm": TemplateField(field_type="BaseLanguageModel", required=True, is_list=False, show=True),
    "func": TemplateField(
        field_type="Callable",
        required=True,
        is_list=False,
        show=True,
        multiline=True,
    ),
    "code": TemplateField(
        field_type="str",
        required=True,
        is_list=False,
        show=True,
        value="",
        multiline=True,
    ),
    "path": TemplateField(
        field_type="file",
        required=True,
        is_list=False,
        show=True,
        value="",
        file_types=[".json", ".yaml", ".yml"],
    ),
}


class ToolCreator(LangChainTypeCreator):
    type_name: str = "tools"
    tools_dict: Optional[Dict] = None

    @property
    def type_to_loader_dict(self) -> Dict:
        settings_service = get_settings_service()
        if self.tools_dict is None:
            all_tools = {}

            for tool, tool_fcn in ALL_TOOLS_NAMES.items():
                try:
                    tool_params = get_tool_params(tool_fcn)
                except Exception:
                    logger.error(f"Error getting params for tool {tool}")
                    continue

                tool_name = tool_params.get("name") or tool

                if tool_name in settings_service.settings.TOOLS or settings_service.settings.DEV:
                    if tool_name == "JsonSpec":
                        tool_params["path"] = tool_params.pop("dict_")  # type: ignore
                    all_tools[tool_name] = {
                        "type": tool,
                        "params": tool_params,
                        "fcn": tool_fcn,
                    }

            self.tools_dict = all_tools

        return self.tools_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        """Get the signature of a tool."""

        base_classes = ["Tool", "BaseTool"]
        fields = []
        params = []
        tool_params = {}

        # Raise error if name is not in tools
        if name not in self.type_to_loader_dict.keys():
            raise ValueError("Tool not found")

        tool_type: str = self.type_to_loader_dict[name]["type"]  # type: ignore

        # if tool_type in _BASE_TOOLS.keys():
        #     params = []
        if tool_type in _LLM_TOOLS.keys():
            params = ["llm"]
        elif tool_type in _EXTRA_LLM_TOOLS.keys():
            extra_keys = _EXTRA_LLM_TOOLS[tool_type][1]
            params = ["llm"] + extra_keys
        elif tool_type in _EXTRA_OPTIONAL_TOOLS.keys():
            extra_keys = _EXTRA_OPTIONAL_TOOLS[tool_type][1]
            params = extra_keys
        # elif tool_type == "Tool":
        #     params = ["name", "description", "func"]
        elif tool_type in CUSTOM_TOOLS:
            # Get custom tool params
            params = self.type_to_loader_dict[name]["params"]  # type: ignore
            base_classes = ["Callable"]
            if node := customs.get_custom_nodes("tools").get(tool_type):
                return node
        elif tool_type in FILE_TOOLS:
            params = self.type_to_loader_dict[name]["params"]  # type: ignore
            base_classes += [name]
        elif tool_type in OTHER_TOOLS:
            tool_dict = build_template_from_class(tool_type, OTHER_TOOLS)
            fields = tool_dict["template"]

            # _type is the only key in fields
            # return None
            if len(fields) == 1 and "_type" in fields:
                return None

            # Pop unnecessary fields and add name
            fields.pop("_type")  # type: ignore
            fields.pop("return_direct", None)  # type: ignore
            fields.pop("verbose", None)  # type: ignore

            tool_params = {
                "name": fields.pop("name")["value"],  # type: ignore
                "description": fields.pop("description")["value"],  # type: ignore
            }

            fields = [
                TemplateField(name=name, field_type=field["type"], **field)
                for name, field in fields.items()  # type: ignore
            ]
            base_classes += tool_dict["base_classes"]

        # Copy the field and add the name
        for param in params:
            field = TOOL_INPUTS.get(param, TOOL_INPUTS["str"]).copy()
            field.name = param
            field.advanced = False
            if param == "aiosession":
                field.show = False
                field.required = False

            fields.append(field)

        template = Template(fields=fields, type_name=tool_type)

        tool_params = {**tool_params, **self.type_to_loader_dict[name]["params"]}
        return {
            "template": util.format_dict(template.to_dict()),
            **tool_params,
            "base_classes": base_classes,
        }

    def to_list(self) -> List[str]:
        """List all load tools"""

        return list(self.type_to_loader_dict.keys())


tool_creator = ToolCreator()



from typing import Optional

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode


class EmbeddingFrontendNode(FrontendNode):
    def add_extra_fields(self) -> None:
        if "VertexAI" in self.template.type_name:
            # Add credentials field which should of type file.
            self.template.add_field(
                TemplateField(
                    field_type="file",
                    required=False,
                    show=True,
                    name="credentials",
                    value="",
                    file_types=[".json"],
                )
            )

    @staticmethod
    def format_vertex_field(field: TemplateField, name: str):
        if "VertexAI" in name:
            key = field.name or ""
            advanced_fields = [
                "verbose",
                "top_p",
                "top_k",
                "max_output_tokens",
            ]
            if key in advanced_fields:
                field.advanced = True
            show_fields = [
                "verbose",
                "project",
                "location",
                "credentials",
                "max_output_tokens",
                "model_name",
                "temperature",
                "top_p",
                "top_k",
            ]

            if key in show_fields:
                field.show = True

    @staticmethod
    def format_jina_fields(field: TemplateField):
        name = field.name or ""
        if "jina" in name:
            field.show = True
            field.advanced = False

        if "auth" in name or "token" in name:
            field.password = True
            field.show = True
            field.advanced = False

        if name == "jina_api_url":
            field.show = True
            field.advanced = True
            field.display_name = "Jina API URL"
            field.password = False

    @staticmethod
    def format_openai_fields(field: TemplateField):
        name = field.name or ""
        if "openai" in name:
            field.show = True
            field.advanced = True
            split_name = name.split("_")
            title_name = " ".join([s.capitalize() for s in split_name])
            field.display_name = title_name.replace("Openai", "OpenAI").replace("Api", "API")

        if "api_key" in name:
            field.password = True
            field.show = True
            field.advanced = False

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        FrontendNode.format_field(field, name)
        if name and "vertex" in name.lower():
            EmbeddingFrontendNode.format_vertex_field(field, name)
        field.advanced = not field.required
        field.show = True
        key = field.name or ""
        if key == "headers":
            field.show = False
        if key == "model_kwargs":
            field.field_type = "dict"
            field.advanced = True
            field.show = True
        elif key in [
            "model_name",
            "temperature",
            "model_file",
            "model_type",
            "deployment_name",
            "credentials",
        ]:
            field.advanced = False
            field.show = True
        if key == "credentials":
            field.field_type = "file"
        if name == "VertexAI" and key not in [
            "callbacks",
            "client",
            "stop",
            "tags",
            "cache",
        ]:
            field.show = True

        # Format Jina fields
        EmbeddingFrontendNode.format_jina_fields(field)
        EmbeddingFrontendNode.format_openai_fields(field)



from typing import List, Optional

from langchain.agents.agent import AgentExecutor
from langchain.agents.agent_toolkits.conversational_retrieval.openai_functions import _get_default_system_message
from langchain.agents.openai_functions_agent.base import OpenAIFunctionsAgent
from langchain.memory.token_buffer import ConversationTokenBufferMemory
from langchain.prompts import SystemMessagePromptTemplate
from langchain.prompts.chat import MessagesPlaceholder
from langchain.schema.memory import BaseMemory
from langchain.tools import Tool
from langchain_community.chat_models import ChatOpenAI
from langflow import CustomComponent
from langflow.field_typing.range_spec import RangeSpec


class ConversationalAgent(CustomComponent):
    display_name: str = "OpenAI Conversational Agent"
    description: str = "Conversational Agent that can use OpenAI's function calling API"

    def build_config(self):
        openai_function_models = [
            "gpt-4-turbo-preview",
            "gpt-4-0125-preview",
            "gpt-4-1106-preview",
            "gpt-4-vision-preview",
            "gpt-3.5-turbo-0125",
            "gpt-3.5-turbo-1106",
        ]
        return {
            "tools": {"display_name": "Tools"},
            "memory": {"display_name": "Memory"},
            "system_message": {"display_name": "System Message"},
            "max_token_limit": {"display_name": "Max Token Limit"},
            "model_name": {
                "display_name": "Model Name",
                "options": openai_function_models,
                "value": openai_function_models[0],
            },
            "code": {"show": False},
            "temperature": {
                "display_name": "Temperature",
                "value": 0.2,
                "range_spec": RangeSpec(min=0, max=2, step=0.1),
            },
        }

    def build(
        self,
        model_name: str,
        openai_api_key: str,
        tools: List[Tool],
        openai_api_base: Optional[str] = None,
        memory: Optional[BaseMemory] = None,
        system_message: Optional[SystemMessagePromptTemplate] = None,
        max_token_limit: int = 2000,
        temperature: float = 0.9,
    ) -> AgentExecutor:
        llm = ChatOpenAI(
            model=model_name,
            api_key=openai_api_key,
            base_url=openai_api_base,
            max_tokens=max_token_limit,
            temperature=temperature,
        )
        if not memory:
            memory_key = "chat_history"
            memory = ConversationTokenBufferMemory(
                memory_key=memory_key,
                return_messages=True,
                output_key="output",
                llm=llm,
                max_token_limit=max_token_limit,
            )
        else:
            memory_key = memory.memory_key  # type: ignore

        _system_message = system_message or _get_default_system_message()
        prompt = OpenAIFunctionsAgent.create_prompt(
            system_message=_system_message,  # type: ignore
            extra_prompt_messages=[MessagesPlaceholder(variable_name=memory_key)],
        )
        agent = OpenAIFunctionsAgent(
            llm=llm,
            tools=tools,
            prompt=prompt,  # type: ignore
        )
        return AgentExecutor(
            agent=agent,
            tools=tools,  # type: ignore
            memory=memory,
            verbose=True,
            return_intermediate_steps=True,
            handle_parsing_errors=True,
        )



from fastapi import APIRouter, HTTPException
from langflow.api.v1.base import (
    Code,
    CodeValidationResponse,
    PromptValidationResponse,
    ValidatePromptRequest,
    validate_prompt,
)
from langflow.template.field.base import TemplateField
from langflow.utils.validate import validate_code
from loguru import logger

# build router
router = APIRouter(prefix="/validate", tags=["Validate"])


@router.post("/code", status_code=200, response_model=CodeValidationResponse)
def post_validate_code(code: Code):
    try:
        errors = validate_code(code.code)
        return CodeValidationResponse(
            imports=errors.get("imports", {}),
            function=errors.get("function", {}),
        )
    except Exception as e:
        return HTTPException(status_code=500, detail=str(e))


@router.post("/prompt", status_code=200, response_model=PromptValidationResponse)
def post_validate_prompt(prompt_request: ValidatePromptRequest):
    try:
        input_variables = validate_prompt(prompt_request.template)
        # Check if frontend_node is None before proceeding to avoid attempting to update a non-existent node.
        if prompt_request.frontend_node is None:
            return PromptValidationResponse(
                input_variables=input_variables,
                frontend_node=None,
            )
        old_custom_fields = get_old_custom_fields(prompt_request)

        add_new_variables_to_template(input_variables, prompt_request)

        remove_old_variables_from_template(old_custom_fields, input_variables, prompt_request)

        update_input_variables_field(input_variables, prompt_request)

        return PromptValidationResponse(
            input_variables=input_variables,
            frontend_node=prompt_request.frontend_node,
        )
    except Exception as e:
        logger.exception(e)
        raise HTTPException(status_code=500, detail=str(e)) from e


def get_old_custom_fields(prompt_request):
    try:
        if len(prompt_request.frontend_node.custom_fields) == 1 and prompt_request.name == "":
            # If there is only one custom field and the name is empty string
            # then we are dealing with the first prompt request after the node was created
            prompt_request.name = list(prompt_request.frontend_node.custom_fields.keys())[0]

        old_custom_fields = prompt_request.frontend_node.custom_fields[prompt_request.name]
        if old_custom_fields is None:
            old_custom_fields = []

        old_custom_fields = old_custom_fields.copy()
    except KeyError:
        old_custom_fields = []
    prompt_request.frontend_node.custom_fields[prompt_request.name] = []
    return old_custom_fields


def add_new_variables_to_template(input_variables, prompt_request):
    for variable in input_variables:
        try:
            template_field = TemplateField(
                name=variable,
                display_name=variable,
                field_type="str",
                show=True,
                advanced=False,
                multiline=True,
                input_types=["Document", "BaseOutputParser"],
                value="",  # Set the value to empty string
            )
            if variable in prompt_request.frontend_node.template:
                # Set the new field with the old value
                template_field.value = prompt_request.frontend_node.template[variable]["value"]

            prompt_request.frontend_node.template[variable] = template_field.to_dict()

            # Check if variable is not already in the list before appending
            if variable not in prompt_request.frontend_node.custom_fields[prompt_request.name]:
                prompt_request.frontend_node.custom_fields[prompt_request.name].append(variable)

        except Exception as exc:
            logger.exception(exc)
            raise HTTPException(status_code=500, detail=str(exc)) from exc


def remove_old_variables_from_template(old_custom_fields, input_variables, prompt_request):
    for variable in old_custom_fields:
        if variable not in input_variables:
            try:
                # Remove the variable from custom_fields associated with the given name
                if variable in prompt_request.frontend_node.custom_fields[prompt_request.name]:
                    prompt_request.frontend_node.custom_fields[prompt_request.name].remove(variable)

                # Remove the variable from the template
                prompt_request.frontend_node.template.pop(variable, None)

            except Exception as exc:
                logger.exception(exc)
                raise HTTPException(status_code=500, detail=str(exc)) from exc


def update_input_variables_field(input_variables, prompt_request):
    if "input_variables" in prompt_request.frontend_node.template:
        prompt_request.frontend_node.template["input_variables"]["value"] = input_variables



from typing import Callable, Union

from langchain_community.utilities.google_search import GoogleSearchAPIWrapper
from langflow import CustomComponent


class GoogleSearchAPIWrapperComponent(CustomComponent):
    display_name = "GoogleSearchAPIWrapper"
    description = "Wrapper for Google Search API."

    def build_config(self):
        return {
            "google_api_key": {"display_name": "Google API Key", "password": True},
            "google_cse_id": {"display_name": "Google CSE ID", "password": True},
        }

    def build(
        self,
        google_api_key: str,
        google_cse_id: str,
    ) -> Union[GoogleSearchAPIWrapper, Callable]:
        return GoogleSearchAPIWrapper(google_api_key=google_api_key, google_cse_id=google_cse_id)  # type: ignore



from langflow.services.task.service import TaskService
from langflow.services.factory import ServiceFactory


class TaskServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(TaskService)

    def create(self):
        # Here you would have logic to create and configure a TaskService
        return TaskService()



import re
from collections import defaultdict
from typing import ClassVar, Dict, List, Optional, Union

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.constants import CLASSES_TO_REMOVE, FORCE_SHOW_FIELDS
from langflow.template.frontend_node.formatter import field_formatters
from langflow.template.template.base import Template
from langflow.utils import constants
from pydantic import BaseModel, Field, field_serializer, model_serializer


class FieldFormatters(BaseModel):
    formatters: ClassVar[Dict] = {
        "openai_api_key": field_formatters.OpenAIAPIKeyFormatter(),
    }
    base_formatters: ClassVar[Dict] = {
        "kwargs": field_formatters.KwargsFormatter(),
        "optional": field_formatters.RemoveOptionalFormatter(),
        "list": field_formatters.ListTypeFormatter(),
        "dict": field_formatters.DictTypeFormatter(),
        "union": field_formatters.UnionTypeFormatter(),
        "multiline": field_formatters.MultilineFieldFormatter(),
        "show": field_formatters.ShowFieldFormatter(),
        "password": field_formatters.PasswordFieldFormatter(),
        "default": field_formatters.DefaultValueFormatter(),
        "headers": field_formatters.HeadersDefaultValueFormatter(),
        "dict_code_file": field_formatters.DictCodeFileFormatter(),
        "model_fields": field_formatters.ModelSpecificFieldFormatter(),
    }

    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        for key, formatter in self.base_formatters.items():
            formatter.format(field, name)

        for key, formatter in self.formatters.items():
            if key == field.name:
                formatter.format(field, name)


class FrontendNode(BaseModel):
    _format_template: bool = True
    template: Template
    description: Optional[str] = None
    icon: Optional[str] = None
    base_classes: List[str]
    name: str = ""
    display_name: Optional[str] = ""
    documentation: str = ""
    custom_fields: Optional[Dict] = defaultdict(list)
    output_types: List[str] = []
    full_path: Optional[str] = None
    field_formatters: FieldFormatters = Field(default_factory=FieldFormatters)

    beta: bool = False
    error: Optional[str] = None

    # field formatters is an instance attribute but it is not used in the class
    # so we need to create a method to get it
    @staticmethod
    def get_field_formatters() -> FieldFormatters:
        return FieldFormatters()

    def set_documentation(self, documentation: str) -> None:
        """Sets the documentation of the frontend node."""
        self.documentation = documentation

    @field_serializer("base_classes")
    def process_base_classes(self, base_classes: List[str]) -> List[str]:
        """Removes unwanted base classes from the list of base classes."""

        return [base_class for base_class in base_classes if base_class not in CLASSES_TO_REMOVE]

    @field_serializer("display_name")
    def process_display_name(self, display_name: str) -> str:
        """Sets the display name of the frontend node."""

        return display_name or self.name

    @model_serializer(mode="wrap")
    def serialize_model(self, handler):
        result = handler(self)
        if hasattr(self, "template") and hasattr(self.template, "to_dict"):
            format_func = self.format_field if self._format_template else None
            result["template"] = self.template.to_dict(format_func)
        name = result.pop("name")

        return {name: result}

    # For backwards compatibility
    def to_dict(self, add_name=True) -> dict:
        """Returns a dict representation of the frontend node."""
        dump = self.model_dump(by_alias=True, exclude_none=True)
        if not add_name:
            return dump.pop(self.name)
        return dump

    def add_extra_fields(self) -> None:
        pass

    def add_extra_base_classes(self) -> None:
        pass

    def add_base_class(self, base_class: Union[str, List[str]]) -> None:
        """Adds a base class to the frontend node."""
        if isinstance(base_class, str):
            self.base_classes.append(base_class)
        elif isinstance(base_class, list):
            self.base_classes.extend(base_class)

    def add_output_type(self, output_type: Union[str, List[str]]) -> None:
        """Adds an output type to the frontend node."""
        if isinstance(output_type, str):
            self.output_types.append(output_type)
        elif isinstance(output_type, list):
            self.output_types.extend(output_type)

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        """Formats a given field based on its attributes and value."""

        FrontendNode.get_field_formatters().format(field, name)

    @staticmethod
    def remove_optional(_type: str) -> str:
        """Removes 'Optional' wrapper from the type if present."""
        return re.sub(r"Optional\[(.*)\]", r"\1", _type)

    @staticmethod
    def check_for_list_type(_type: str) -> tuple:
        """Checks for list type and returns the modified type and a boolean indicating if it's a list."""
        is_list = "List" in _type or "Sequence" in _type
        if is_list:
            _type = re.sub(r"(List|Sequence)\[(.*)\]", r"\2", _type)
        return _type, is_list

    @staticmethod
    def replace_mapping_with_dict(_type: str) -> str:
        """Replaces 'Mapping' with 'dict'."""
        return _type.replace("Mapping", "dict")

    @staticmethod
    def handle_union_type(_type: str) -> str:
        """Simplifies the 'Union' type to the first type in the Union."""
        if "Union" in _type:
            _type = _type.replace("Union[", "")[:-1]
            _type = _type.split(",")[0]
            _type = _type.replace("]", "").replace("[", "")
        return _type

    @staticmethod
    def handle_special_field(field, key: str, _type: str, SPECIAL_FIELD_HANDLERS) -> str:
        """Handles special field by using the respective handler if present."""
        handler = SPECIAL_FIELD_HANDLERS.get(key)
        return handler(field) if handler else _type

    @staticmethod
    def handle_dict_type(field: TemplateField, _type: str) -> str:
        """Handles 'dict' type by replacing it with 'code' or 'file' based on the field name."""
        if "dict" in _type.lower() and field.name == "dict_":
            field.field_type = "file"
            field.file_types = [".json", ".yaml", ".yml"]
        elif _type.startswith("Dict") or _type.startswith("Mapping") or _type.startswith("dict"):
            field.field_type = "dict"
        return _type

    @staticmethod
    def replace_default_value(field: TemplateField, value: dict) -> None:
        """Replaces default value with actual value if 'default' is present in value."""
        if "default" in value:
            field.value = value["default"]

    @staticmethod
    def handle_specific_field_values(field: TemplateField, key: str, name: Optional[str] = None) -> None:
        """Handles specific field values for certain fields."""
        if key == "headers":
            field.value = """{"Authorization": "Bearer <token>"}"""
        FrontendNode._handle_model_specific_field_values(field, key, name)
        FrontendNode._handle_api_key_specific_field_values(field, key, name)

    @staticmethod
    def _handle_model_specific_field_values(field: TemplateField, key: str, name: Optional[str] = None) -> None:
        """Handles specific field values related to models."""
        model_dict = {
            "OpenAI": constants.OPENAI_MODELS,
            "ChatOpenAI": constants.CHAT_OPENAI_MODELS,
            "Anthropic": constants.ANTHROPIC_MODELS,
            "ChatAnthropic": constants.ANTHROPIC_MODELS,
        }
        if name in model_dict and key == "model_name":
            field.options = model_dict[name]
            field.is_list = True

    @staticmethod
    def _handle_api_key_specific_field_values(field: TemplateField, key: str, name: Optional[str] = None) -> None:
        """Handles specific field values related to API keys."""
        if "api_key" in key and "OpenAI" in str(name):
            field.display_name = "OpenAI API Key"
            field.required = False
            if field.value is None:
                field.value = ""

    @staticmethod
    def handle_kwargs_field(field: TemplateField) -> None:
        """Handles kwargs field by setting certain attributes."""

        if "kwargs" in (field.name or "").lower():
            field.advanced = True
            field.required = False
            field.show = False

    @staticmethod
    def handle_api_key_field(field: TemplateField, key: str) -> None:
        """Handles api key field by setting certain attributes."""
        if "api" in key.lower() and "key" in key.lower():
            field.required = False
            field.advanced = False

            field.display_name = key.replace("_", " ").title()
            field.display_name = field.display_name.replace("Api", "API")

    @staticmethod
    def should_show_field(key: str, required: bool) -> bool:
        """Determines whether the field should be shown."""
        return (
            (required and key not in ["input_variables"])
            or key in FORCE_SHOW_FIELDS
            or "api" in key
            or ("key" in key and "input" not in key and "output" not in key)
        )

    @staticmethod
    def should_be_password(key: str, show: bool) -> bool:
        """Determines whether the field should be a password field."""
        return any(text in key.lower() for text in {"password", "token", "api", "key"}) and show

    @staticmethod
    def should_be_multiline(key: str) -> bool:
        """Determines whether the field should be multiline."""
        return key in {
            "suffix",
            "prefix",
            "template",
            "examples",
            "code",
            "headers",
            "description",
        }

    @staticmethod
    def set_field_default_value(field: TemplateField, value: dict, key: str) -> None:
        """Sets the field value with the default value if present."""
        if "default" in value:
            field.value = value["default"]
        if key == "headers":
            field.value = """{"Authorization": "Bearer <token>"}"""



from typing import Optional, Union

from langchain.embeddings.base import Embeddings
from langchain_community.vectorstores import VectorStore
from langchain_community.vectorstores.pgvector import PGVector
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever
from langflow import CustomComponent


class PGVectorComponent(CustomComponent):
    """
    A custom component for implementing a Vector Store using PostgreSQL.
    """

    display_name: str = "PGVector"
    description: str = "Implementation of Vector Store using PostgreSQL"
    documentation = "https://python.langchain.com/docs/integrations/vectorstores/pgvector"

    def build_config(self):
        """
        Builds the configuration for the component.

        Returns:
        - dict: A dictionary containing the configuration options for the component.
        """
        return {
            "code": {"show": False},
            "documents": {"display_name": "Documents", "is_list": True},
            "embedding": {"display_name": "Embedding"},
            "pg_server_url": {
                "display_name": "PostgreSQL Server Connection String",
                "advanced": False,
            },
            "collection_name": {"display_name": "Table", "advanced": False},
        }

    def build(
        self,
        embedding: Embeddings,
        pg_server_url: str,
        collection_name: str,
        documents: Optional[Document] = None,
    ) -> Union[VectorStore, BaseRetriever]:
        """
        Builds the Vector Store or BaseRetriever object.

        Args:
        - embedding (Embeddings): The embeddings to use for the Vector Store.
        - documents (Optional[Document]): The documents to use for the Vector Store.
        - collection_name (str): The name of the PG table.
        - pg_server_url (str): The URL for the PG server.

        Returns:
        - VectorStore: The Vector Store object.
        """

        try:
            if documents is None:
                vector_store = PGVector.from_existing_index(
                    embedding=embedding,
                    collection_name=collection_name,
                    connection_string=pg_server_url,
                )
            else:
                vector_store = PGVector.from_documents(
                    embedding=embedding,
                    documents=documents,  # type: ignore
                    collection_name=collection_name,
                    connection_string=pg_server_url,
                )
        except Exception as e:
            raise RuntimeError(f"Failed to build PGVector: {e}")
        return vector_store



from langflow.template.frontend_node import (
    agents,
    chains,
    embeddings,
    llms,
    memories,
    prompts,
    tools,
    vectorstores,
    documentloaders,
    textsplitters,
    custom_components,
)

__all__ = [
    "agents",
    "chains",
    "embeddings",
    "memories",
    "tools",
    "llms",
    "prompts",
    "vectorstores",
    "documentloaders",
    "textsplitters",
    "custom_components",
]



import os
from typing import List, Optional, Union

import pinecone  # type: ignore
from langchain.schema import BaseRetriever
from langchain_community.vectorstores import VectorStore
from langchain_community.vectorstores.pinecone import Pinecone
from langflow import CustomComponent
from langflow.field_typing import Document, Embeddings


class PineconeComponent(CustomComponent):
    display_name = "Pinecone"
    description = "Construct Pinecone wrapper from raw documents."

    def build_config(self):
        return {
            "documents": {"display_name": "Documents"},
            "embedding": {"display_name": "Embedding"},
            "index_name": {"display_name": "Index Name"},
            "namespace": {"display_name": "Namespace"},
            "pinecone_api_key": {"display_name": "Pinecone API Key", "default": "", "password": True, "required": True},
            "pinecone_env": {"display_name": "Pinecone Environment", "default": "", "required": True},
            "search_kwargs": {"display_name": "Search Kwargs", "default": "{}"},
            "pool_threads": {"display_name": "Pool Threads", "default": 1, "advanced": True},
        }

    def build(
        self,
        embedding: Embeddings,
        pinecone_env: str,
        documents: List[Document],
        text_key: str = "text",
        pool_threads: int = 4,
        index_name: Optional[str] = None,
        pinecone_api_key: Optional[str] = None,
        namespace: Optional[str] = "default",
    ) -> Union[VectorStore, Pinecone, BaseRetriever]:
        if pinecone_api_key is None or pinecone_env is None:
            raise ValueError("Pinecone API Key and Environment are required.")
        if os.getenv("PINECONE_API_KEY") is None and pinecone_api_key is None:
            raise ValueError("Pinecone API Key is required.")

        pinecone.init(api_key=pinecone_api_key, environment=pinecone_env)  # type: ignore
        if not index_name:
            raise ValueError("Index Name is required.")
        if documents:
            return Pinecone.from_documents(
                documents=documents,
                embedding=embedding,
                index_name=index_name,
                pool_threads=pool_threads,
                namespace=namespace,
                text_key=text_key,
            )

        return Pinecone.from_existing_index(
            index_name=index_name,
            embedding=embedding,
            text_key=text_key,
            namespace=namespace,
            pool_threads=pool_threads,
        )
